################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
"""Wrapper for TACE model in TorchSim.

Based on https://github.com/TorchSim/torch-sim/blob/main/torch_sim/models

This module provides a TorchSim wrapper of the TACE model for computing
energies, forces, and stresses for atomistic systems. It integrates the TACE model
with TorchSim's simulation framework, handling batched computations for multiple
systems simultaneously.
"""

from collections.abc import Callable
from pathlib import Path
from typing import Optional

import torch
import torch_sim as ts
from torch_sim.models.interface import ModelInterface
from torch_sim.neighbors import vesin_nl_ts
from torch_sim.typing import StateDict
from tace.lightning import load_tace
from tace.dataset.element import TorchElement
from tace.utils._global import DTYPE, DEVICE

class TACETorchSimCalc(ModelInterface):
    def __init__(
        self,
        model: str | Path | torch.nn.Module | None = None,
        *,
        level: Optional[int] = None,
        spin_on: Optional[bool] = None,
        target_property: Optional[list[str]] = None,
        device: torch.device | None = None,
        dtype: torch.dtype = torch.float64, 
        neighbor_list_fn: Callable = vesin_nl_ts,
        compute_forces: bool = True,
        compute_stress: bool = True,
        atomic_numbers: torch.Tensor | None = None,
        system_idx: torch.Tensor | None = None,
    ) -> None:
        """
        The model can be initialized with atomic numbers
        and system indices, or these can be provided during the forward pass.

        Args:
            level : int
                Specify which fidelity level to use. 
            spin_on : bool
                If your model uses spin_on uie embedding, you can control whether 
                your calculation enables spin polarization.
            target_property: list(str)
                Extra caculate hessians, atomic_virials, Conservative polarizability, etc,
                If you want to use this parameter, you must provide all the required physical quantities.
            atomic_numbers (torch.Tensor | None): Atomic numbers with shape [n_atoms].
                If provided at initialization, cannot be provided again during forward.
            system_idx (torch.Tensor | None): System indices with shape [n_atoms]
                indicating which system each atom belongs to. If not provided with
                atomic_numbers, all atoms are assumed to be in the same system.
        """
        super().__init__()
        self._device = DEVICE[device] or torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )
        
        self._dtype = DTYPE[dtype]
        self.neighbor_list_fn = neighbor_list_fn
        self._compute_forces = compute_forces
        self._compute_stress = compute_stress
        self._memory_scales_with = "n_atoms_x_density"

        # Load TACE model
        model = load_tace(
            model, 
            self._device, 
            strict=True, 
            use_ema=True, 
            target_property=target_property
        ) 
        if level is not None:
            self.level = level
            model.reset_computing_level(level) 
        else:
            self.level = model.get_computing_level()
        if spin_on is not None:
            self.spin_on = spin_on
            model.reset_spin_on(spin_on) 
        else:
            self.spin_on = model.get_spin_on() 
        model.eval()
        for param in model.parameters():
            param.requires_grad = False
        self.model = model
        if self.dtype is not None:
            self.model = self.model.to(dtype=self.dtype)

        # Set model properties
        self.r_max = self.model.readout_fn.cutoff
        self.torch_element = TorchElement(
            self.model.readout_fn.atomic_numbers.cpu().tolist()
        )

        # Store flag to track if atomic numbers were provided at init
        self.atomic_numbers_in_init = atomic_numbers is not None

        # Set up system_idx information if atomic numbers are provided
        if atomic_numbers is not None:
            if system_idx is None:
                # If system_idx is not provided, assume all atoms belong to same system
                system_idx = torch.zeros(
                    len(atomic_numbers), dtype=torch.long, device=self.device
                )

            self.setup_from_system_idx(atomic_numbers, system_idx)

        if self._compute_forces:
            self.model.flags.compute_forces = True
            self.model.compute_first_derivative = True
        if self._compute_stress:
            self.model.flags.compute_stress = True
            self.model.compute_first_derivative = True

    def setup_from_system_idx(
        self, atomic_numbers: torch.Tensor, system_idx: torch.Tensor
    ) -> None:
        self.atomic_numbers = atomic_numbers
        self.system_idx = system_idx

        # Determine number of systems and atoms per system
        self.n_systems = system_idx.max().item() + 1

        # Create ptr tensor for system boundaries
        self.n_atoms_per_system = []
        ptr = [0]
        for sys_idx in range(self.n_systems):
            system_mask = system_idx == sys_idx
            n_atoms = system_mask.sum().item()
            self.n_atoms_per_system.append(n_atoms)
            ptr.append(ptr[-1] + n_atoms)

        self.ptr = torch.tensor(ptr, dtype=torch.long, device=self.device)
        self.total_atoms = atomic_numbers.shape[0]
        self.node_attrs = self.torch_element.z2onehot(atomic_numbers).to(self._dtype)

    # onehot = element.z2onehot(atomic_numbers).to(dtype=torch.get_default_dtype())

    def forward(self, state: ts.SimState | StateDict) -> dict[str, torch.Tensor]:
        sim_state = (
            state
            if isinstance(state, ts.SimState)
            else ts.SimState(**state, masses=torch.ones_like(state["positions"]))
        )

        # Handle input validation for atomic numbers
        if sim_state.atomic_numbers is None and not self.atomic_numbers_in_init:
            raise ValueError(
                "Atomic numbers must be provided in either the constructor or forward."
            )
        if sim_state.atomic_numbers is not None and self.atomic_numbers_in_init:
            raise ValueError(
                "Atomic numbers cannot be provided in both the constructor and forward."
            )

        # Use system_idx from init if not provided
        if sim_state.system_idx is None:
            if not hasattr(self, "system_idx"):
                raise ValueError(
                    "System indices must be provided if not set during initialization"
                )
            sim_state.system_idx = self.system_idx

        # Update system_idx information if new atomic numbers are provided
        if (
            sim_state.atomic_numbers is not None
            and not self.atomic_numbers_in_init
            and not torch.equal(
                sim_state.atomic_numbers,
                getattr(self, "atomic_numbers", torch.zeros(0, device=self.device)),
            )
        ):
            self.setup_from_system_idx(sim_state.atomic_numbers, sim_state.system_idx)

        # Process each system's neighbor list separately
        edge_indices = []
        edge_shifts_list = []
        offset = 0

        # TODO (AG): Currently doesn't work for batched neighbor lists
        for sys_idx in range(self.n_systems):
            system_mask = sim_state.system_idx == sys_idx
            # Calculate neighbor list for this system
            edge_idx, shifts_idx = self.neighbor_list_fn(
                positions=sim_state.positions[system_mask],
                cell=sim_state.row_vector_cell[sys_idx],
                pbc=sim_state.pbc,
                cutoff=self.r_max,
            )

            # Adjust indices for the system
            edge_idx = edge_idx + offset

            edge_indices.append(edge_idx)
            edge_shifts_list.append(shifts_idx)

            offset += len(sim_state.positions[system_mask])

        # Combine all neighbor lists
        edge_index = torch.cat(edge_indices, dim=1)
        edge_shifts = torch.cat(edge_shifts_list, dim=0)

        # Get model output
        out = self.model(
            dict(
                ptr=self.ptr,
                batch=sim_state.system_idx,
                node_attrs=self.node_attrs,
                # pbc=sim_state.pbc,
                lattice=sim_state.row_vector_cell,
                positions=sim_state.positions,
                edge_index=edge_index,
                edge_shifts=edge_shifts,
            ),
        )

        results: dict[str, torch.Tensor] = {}

        # Process energy
        energy = out["energy"]
        if energy is not None:
            results["energy"] = energy.detach()
        else:
            results["energy"] = torch.zeros(self.n_systems, device=self.device)

        # Process forces
        if self.compute_forces:
            forces = out["forces"]
            if forces is not None:
                results["forces"] = forces.detach()

        # Process stress
        if self.compute_stress:
            stress = out["stress"]
            if stress is not None:
                results["stress"] = stress.detach()

        return results


################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import re
import yaml
import argparse
from pathlib import Path


import torch


from ..lightning import load_tace
from ..dataset.element import atomic_numbers


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-m", "--model",
        type=str,
        required=True,
        help="Model path, *.ckpt, *.pt, *.pth",
    )
    parser.add_argument(
        "-u", "--update",
        type=str,
        nargs='+',  
        choices=["atomic_energy", 'scale', 'shift'],
        # choices=["atomic_energy", 'avg_num_neighbors', 'scale', 'shift']
        default="atomic_energy",
        help="specify statistics info to update",
    )
    return parser.parse_args()

def safe_dict(info: dict):
    new_dict = {}
    for k, v in info.items():
        if k in atomic_numbers:
            k = atomic_numbers[k]
        new_dict[k] = v
    return dict(sorted(new_dict.items()))

def main():
    args = parse_args()
    model = load_tace(args.model, device='cpu', strict=True, use_ema=1)

    pattern = re.compile(r"statistics_(\d+)\.yaml")
    statistics_dict = {}
    for path in Path(".").iterdir():
        if path.is_file():
            m = pattern.fullmatch(path.name)
            if m:
                level = int(m.group(1))
                statistics_dict[level] = yaml.safe_load(path.read_text())
    statistics_dict: dict = dict(sorted(statistics_dict.items()))
    with torch.no_grad():
        if 'atomic_energy' in args.update:
            atomic_energies = model.readout_fn.atomic_energy_layer.atomic_energy.detach().clone()
            for level, stats in statistics_dict.items():
                if "atomic_energy" not in stats:
                    continue
                new_row = torch.tensor(
                    [v for _, v in safe_dict(stats["atomic_energy"]).items()],
                    dtype=atomic_energies.dtype,
                    device=atomic_energies.device,
                )
                atomic_energies[level, :] = new_row
            model.readout_fn.atomic_energy_layer.atomic_energy.copy_(atomic_energies)

        if 'scale' in args.update:
            scales = model.readout_fn.scale_shift.scale.detach().clone()
            for level, stats in statistics_dict.items():
                if "scale" not in stats:
                    continue
                new_row = torch.tensor(
                    [v for _, v in safe_dict(stats["scale"]).items()],
                    dtype=scales.dtype,
                    device=scales.device,
                )
                scales[level, :] = new_row
            model.readout_fn.scale_shift.scale.copy_(scales)
        
        if 'shift' in args.update:
            shifts = model.readout_fn.scale_shift.shift.detach().clone()
            for level, stats in statistics_dict.items():
                if "shift" not in stats:
                    continue
                new_row = torch.tensor(
                    [v for _, v in safe_dict(stats["shift"]).items()],
                    dtype=shifts.dtype,
                    device=shifts.device,
                )
                shifts[level, :] = new_row
            model.readout_fn.scale_shift.shift.copy_(shifts)
    print(model)
    torch.save(model, "new_statistics.pt")





################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import yaml
import logging
import warnings
from pathlib import Path


import hydra
from omegaconf import DictConfig, OmegaConf


from ..dataset.statistics import Statistics
from ..lightning.trainer import train
from ..lightning.lit_model import finetune, load_tace
from ..lightning.select_model import select_model
from ..dataset.dataloader import build_atomsList, compute_statistics
from ..dataset.datamodule import build_datamodule
from ..utils.hydra_resolver import register_resolvers
from ..utils.logger import set_logger
from ..utils.utils import (
    set_global_seed,
    set_precision,
    save_full_cfg,
    deep_convert,
)
from ..utils.env import set_env
from ..dataset.quantity import (
    KEYS,
    KeySpecification,
    update_keyspec_from_kwargs,
    get_target_property,
)
from ..dataset.quantity import get_embedding_property


register_resolvers()


def initialize(cfg):
    cfg = deep_convert(cfg)
    set_logger()
    if cfg['misc'].get('ignore_warning', True): 
        try:
            warnings.simplefilter("ignore", FutureWarning)
            warnings.filterwarnings(
                "ignore", module="pydantic._internal._generate_schema"
            )
        except Exception:
            pass
    save_full_cfg(cfg)
    set_env(cfg)
    set_global_seed(cfg)
    set_precision(cfg)
    return cfg


def build(cfg: DictConfig):
    cfg = initialize(OmegaConf.to_container(cfg, resolve=True, structured_config_mode="dict"))
    target_property = get_target_property(cfg)
    embedding_property = get_embedding_property(cfg)
    userKeys = KEYS
    userKeys.update(cfg['dataset'].get('keys', {}))
    keyspec = KeySpecification()
    update_keyspec_from_kwargs(keyspec, userKeys)
    num_levels = cfg['model']['config'].get("num_levels", 1)

    statistics = None
    # finetune or reusme, statistics is no need to recalculate
    if not (cfg.get("finetune_from_model", None) or cfg.get("resume_from_model", None)): 
        statistics_yaml = [Path('.') / f'statistics_{i}.yaml' for i in range(num_levels)]
        if all(yaml_file.exists() for yaml_file in statistics_yaml):
            statistics = []
            for yaml_file in statistics_yaml:
                with open(yaml_file, "r") as f:
                    statistics_data = yaml.safe_load(f)
                    statistics.append(Statistics(**statistics_data))
            for idx, yaml_file in enumerate(statistics_yaml):
                logging.info(f"Using statistics_yaml from '{str(yaml_file)}' for level {idx}")
        else:
            logging.info(f"Computing statistics information from scratch")
            element, threeAtomsList, atomic_energies = build_atomsList(
                cfg, target_property, embedding_property, keyspec, num_levels
            )
            statistics = compute_statistics(
                cfg, 
                target_property, 
                embedding_property, 
                keyspec, 
                num_levels,
                element,
                threeAtomsList,
                atomic_energies,
            )

    if cfg.get("finetune_from_model", None):
        model = finetune(cfg)
        cfg['model']['config'] = model.readout_fn.model_config
        statistics = model.readout_fn.statistics
        finetune_cfg = cfg.get('finetune', {})
        if finetune_cfg:
            logging.info(f"finetune_config field found in your main train config, use lora and parameter freezing")
        else:
            yaml_path = Path("finetune_config.yaml")
            if yaml_path.exists():
                logging.info(f"{yaml_path} found, use lora and parameter freezing")
                try:
                    finetune_cfg = yaml.safe_load(yaml_path.read_text())
                except Exception as e:
                    logging.error(f"Failed to read {yaml_path}: {e}")
            else:
                logging.warning(f"{yaml_path} not found, skipping lora and parameter freezing.")
        cfg['finetune'] = finetune_cfg
    elif cfg.get("resume_from_model", None):
        model = load_tace(
                cfg["resume_from_model"],
                device="cpu",
                strict=True,
                use_ema=True,
            )
        cfg['model']['config'] = model.readout_fn.model_config
        statistics = model.readout_fn.statistics
        cfg['finetune'] = cfg.get('finetune', {})

    else: # finetune_from_model or from scratch
        model = select_model(cfg, statistics, target_property, embedding_property)

    datamodule = build_datamodule(cfg, target_property, embedding_property, keyspec, num_levels, statistics)

    return cfg, statistics, target_property, embedding_property, model, datamodule

@hydra.main(version_base="1.3", config_path=str(Path.cwd()), config_name="tace")
def main(cfg: DictConfig):
    cfg, statistics, target_property, embedding_property, model, datamodule = build(cfg)

    train_arguments = {
        "cfg": cfg,
        "statistics": statistics,
        "target_property": target_property,
        "embedding_property": embedding_property,
        "model": model,
        "datamodule": datamodule,
    }
    train(**train_arguments)


if __name__ == "__main__":
    main()




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import List, Optional


import torch
from torch import nn
from omegaconf import ListConfig


from .fn import LOSS_FN


class UncertaintyLoss(nn.Module):
    """
    Multi-task uncertainty weighted loss.
    Based on: "Multi-Task Learning Using Uncertainty to Weigh Losses for Scene Geometry and Semantics"
    (https://arxiv.org/abs/1705.07115)
    """

    def __init__(
        self,
        loss_property: List[str],
        loss_function_name: List[str],
        loss_property_weights: Optional[List[float]] = None,
        **kwargs,
    ):
        super().__init__()
        init_log_sigmas = loss_property_weights
        assert isinstance(
            loss_property, (List, ListConfig)
        ), f"``cfg.loss.loss_property`` should be a list, got {type(loss_property)}"
        assert isinstance(
            loss_function_name, (List, ListConfig)
        ), f"``cfg.loss.loss_function_name`` should be a list, got {type(loss_property)}"
        if init_log_sigmas is None:
            init_log_sigmas = [0.0] * len(loss_property)
        assert isinstance(
            init_log_sigmas, (List, ListConfig)
        ), f"``cfg.loss.loss_property_weights`` should be a list, got {type(loss_property)}"
        assert len(loss_property) == len(loss_function_name) == len(init_log_sigmas)
        for fn in loss_function_name:
            assert (
                fn in LOSS_FN
            ), f"{fn} not in LOSS_FN, add by yourself, all available function name are {list(LOSS_FN)}"

        self.loss_property = loss_property
        self.loss_function_name = loss_function_name
        self.log_sigmas = nn.ParameterDict()
        for p, val in zip(loss_property, init_log_sigmas):
            self.log_sigmas[p] = nn.Parameter(
                torch.tensor(val, dtype=torch.get_default_dtype())
            )

    def forward(self, pred, label):
        total_loss = 0.0
        for p, fn_name in zip(self.loss_property, self.loss_function_name):
            p_loss = LOSS_FN[fn_name](pred, label)
            log_sigma = self.log_sigmas[p]
            total_loss += 0.5 * torch.exp(-log_sigma) * p_loss + log_sigma
        return total_loss

    def __repr__(self):
        task_strs = [
            f"{p} (fn={fn}, log_sigma={v.item():.4f})"
            for p, fn, v in zip(
                self.loss_property, self.loss_function_name, self.log_sigmas.values()
            )
        ]
        tasks_info = ", ".join(task_strs)
        return f"{self.__class__.__name__}({tasks_info})"



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import random
import argparse


from ase import Atoms
from ase.io import read, write


def parse_args():
    parser = argparse.ArgumentParser(
        description="Split dataset into train/val/test sets with info['xzm_index'] labeling"
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        required=True,
        help="Path to the structure file, e.g. data.traj or data.xyz",
    )
    parser.add_argument(
        "-n",
        "--num",
        type=int,
        nargs=3,
        metavar=("TRAIN", "VAL", "TEST"),
        required=True,
        help="Three non-negative integers: number of train, val, and test samples",
    )
    parser.add_argument("-s", "--seed", type=int, default=42, help="Random seed")
    parser.add_argument("--train_index", type=str, default=None, help="Split from index file (index start from zero)")
    parser.add_argument("--valid_index", type=str, default=None, help="Split from index file (index start from zero)")
    parser.add_argument("--test_index", type=str, default=None, help="Split from index file (index start from zero)")
    return parser.parse_args()


def main():

    args = parse_args()
    structures: list[Atoms] = read(args.input, index=":")
    total = len(structures)
    print(f"Total structures loaded: {total}")

    # Requested split sizes
    n_train, n_val, n_test = args.num
    total_requested = n_train + n_val + n_test

    if total_requested == 0:
        print("Nothing to split: all target sizes are 0.")
        return

    if total_requested > total:
        raise ValueError(
            f"Requested total of {total_requested} structures "
            f"(train={n_train}, val={n_val}, test={n_test}), "
            f"but only {total} structures are available in the file."
        )

    if total_requested == total:
        print(f"All structures in {args.input} have benn used.")

    print(f"Final split sizes: train={n_train}, val={n_val}, test={n_test}")

    # Shuffle and assign indices
    random.seed(args.seed)
    indices = list(range(total))
    random.shuffle(indices)

    split_tags = ["train"] * n_train + ["val"] * n_val + ["test"] * n_test

    tag_dict = {"train": [], "val": [], "test": []}

    for i, idx in enumerate(indices[: len(split_tags)]):
        tag = split_tags[i]
        atoms = structures[idx]
        atoms.info["tace_index"] = idx  # <-- original index
        tag_dict[tag].append(atoms)

    base_name = args.input.rsplit(".", 1)[0]

    for tag, atoms_list in tag_dict.items():
        if atoms_list:
            out_file = f"{base_name}_{tag}.xyz"
            write(out_file, atoms_list)
            print(f"Wrote {len(atoms_list)} structures to {out_file}")


if __name__ == "__main__":
    main()



# import tomli
# import importlib.resources

# def get_version():
#     with importlib.resources.open_text("tace", "pyproject.toml") as f:
#         pyproject = tomli.load(f)
#     return pyproject["project"]["version"]

# __version__ = get_version()


__version__ = "0.0.8"




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from pathlib import Path
from typing import List, Sequence, Tuple


import numpy as np


def random_split(
    train_list: Sequence, valid_ratio: float, seed: int
) -> Tuple[List, List]:
    '''
    The reason why both indices need to be saved is that when I tried to 
    reproduce my own results, I found that only loading the valid indices
    would lead to different training results. This happens because the order of 
    structures in the training set changes during loading, which in turn alters 
    the order of batches.
    '''
    size = len(train_list)
    train_size = size - int(valid_ratio * size)

    indices = list(range(size))
    rng = np.random.default_rng(seed)
    rng.shuffle(indices)

    train_index_file = Path(".") / "train.index"  # index starts from zero
    valid_index_file = Path(".") / "valid.index"  # index starts from zero
    with open(train_index_file, "w") as f:
        for idx in indices[:train_size]: # not allowed sorted
            f.write(f"{idx}\n")
    with open(valid_index_file, "w") as f:
        for idx in indices[train_size:]:
            f.write(f"{idx}\n")

    return (
        [train_list[i] for i in indices[:train_size]],
        [train_list[i] for i in indices[train_size:]],
    )



import torch
from typing import Sequence, Dict

class ElementAwareMixin:
    def _init_elements(self, atomic_numbers: Sequence[int]):
        zs = sorted(list(set(int(z) for z in atomic_numbers)))
        self.atomic_numbers = zs
        self.num_elements = len(zs)
        # z -> element index
        self._z_to_idx: Dict[int, int] = {
            z: i for i, z in enumerate(zs)
        }

    def get_element_parameter(
            self, 
            param: torch.Tensor, 
            atomic_numbers: Sequence[int],
        ) -> torch.Tensor:

        param_list = []
        for z in atomic_numbers:
            param_list.append(param[self._z_to_idx[z]])

        return torch.stack(param_list, dim=0)



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################


from typing import Dict
from itertools import combinations


import torch
from torch import Tensor


from .ictd import ICTD
from .utils import expand_dims_to


def factorial(n: int) -> int:
    result = 1
    for i in range(2, n + 1):
        result *= i
    return result


def double_factorial(n: int) -> int:
    result = 1
    for i in range(n, 0, -2):
        result *= i
    return result


def _norm(n: int) -> float:
    """(2n - 1)!! / n!"""
    num = double_factorial(2 * n - 1)
    den = factorial(n)
    return num / den


def delta_tensor(i: int, j: int, ndim: int, device=None, dtype=None) -> Tensor:
    delta = torch.eye(3, device=device, dtype=dtype)
    for _ in range(ndim - 2):
        delta = delta.unsqueeze(0)
    perm = list(range(ndim))
    perm[i], perm[-2] = perm[-2], perm[i]
    perm[j], perm[-1] = perm[-1], perm[j]
    delta = delta.permute(*perm)
    return delta


def symmetric_outer_product(v: Tensor, n: int, norm: bool = True) -> Tensor:
    out = torch.ones_like(v[..., 0])
    for _ in range(n):
        out = out[..., None] * expand_dims_to(v, out.ndim + 1, dim=v.ndim - 1)
    if norm:
        out = out * _norm(n)
    return out


def subtract_traces(T: Tensor, n: int) -> Tensor:
    result = T.clone()
    base_combs = list(combinations(range(-n, 0), 2))
    for k in range(1, n // 2 + 1):
        denom = 1.0
        for j in range(2, k + 2):
            denom *= 3 + 2 * (n - j)
        coeff = ((-1) ** k) / denom
        corr = torch.zeros_like(T)
        for pairs in combinations(base_combs, k):
            idxs = [idx for pair in pairs for idx in pair]
            if len(set(idxs)) < 2 * k:
                continue
            delta = torch.ones_like(T)
            for i, j in pairs:
                delta = delta * delta_tensor(
                    i, j, n + 1, device=T.device, dtype=T.dtype
                )
            trace = torch.sum(T * delta, dim=tuple(idxs), keepdim=True)
            corr += delta * trace
        result = result + coeff * corr
    return result

# numerical
def symmetric_traceless_outer_product(v: Tensor, n: int, norm: bool = True) -> Tensor:
    T = symmetric_outer_product(v, n, norm)
    return subtract_traces(T, n)


class LegacyCartesianHarmonics1(torch.nn.Module):
    def __init__(self, lmax: int, norm: bool = True, traceless: bool = True) -> None:
        super().__init__()
        self.lmax = lmax
        self.norm = norm
        self.traceless = traceless
        PS, DS, CS, SS = ICTD(lmax)
        self.register_buffer(f"Q", DS[0].to(torch.get_default_dtype()))
        del PS, DS, CS, SS

    def forward(self, v: Tensor) -> Tensor:
        T = symmetric_outer_product(v, self.lmax, self.norm)
        if self.traceless:
            B = T.size(0)
            if B == 0:
                return T 
            else:
                REST = T.size()[1:]
                T = T.reshape(B, -1)
                T = T @ self.Q
                T = T.reshape((B,) + REST)
                return T
        else:
            return T

    def __repr__(self):
        return f"{self.__class__.__name__}(r={self.lmax}, norm={self.norm})"
    

class LegacyCartesianHarmonics2(torch.nn.Module):
    def __init__(self, lmax: int, norm: bool = True, traceless: bool = True) -> None:
        super().__init__()
        self.lmax = lmax
        self.norm = norm
        self.traceless = traceless
        for l in range(self.lmax+1):
            PS, DS, CS, SS = ICTD(l)
            self.register_buffer(f"D{l}", DS[0].to(torch.get_default_dtype()))
            del PS, DS, CS, SS

    def forward(self, v: Tensor) -> Dict[int, Tensor]:
        T = torch.ones_like(v[..., 0])
        edge_attrs: Dict[int, Tensor] = {}
        edge_attrs[0] = T.unsqueeze(1)

        for l in range(1, self.lmax+1):
            T = T[..., None] * expand_dims_to(v, T.ndim + 1, dim=v.ndim - 1)
            edge_attrs[l] = T
                
        if self.traceless:
            for l in range(1, self.lmax+1):
                T = edge_attrs[l]
                B = T.size(0)
                if B != 0:
                    REST = T.size()[1:]
                    T = T.reshape(B, -1)
                    T = T @ self.D(l)
                    T = T.reshape((B,) + REST)
                edge_attrs[l] = T

        if self.norm:
            for l in range(1, self.lmax+1):
                edge_attrs[l] = edge_attrs[l] * _norm(l)
                
        for l in range(1, self.lmax+1):
            edge_attrs[l] = edge_attrs[l].unsqueeze(1)

        return edge_attrs

    def D(self, l: int):
        return dict(self.named_buffers())[f"D{l}"]
    
    def __repr__(self):
        return f"{self.__class__.__name__}(r={self.lmax}, norm={self.norm}, traceless={self.traceless})"
    



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
"""
NormGate are Modified from HotPP.
https://gitlab.com/bigd4/hotpp

MIT License

Copyright (c) 2024 wang laosi

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import Dict
import torch
import torch.nn.functional as F


from .act import ACT
from .utils import expand_dims_to

class NormGateBase(torch.nn.Module):
    """
    Activate function for tensor inputs with shape [batch, channel] + [3,] * l
    """
    r: int
    def __init__(
        self,
        r: int,
        in_dim: int,
    ) -> None:
        super().__init__()

        self.r = r

        if self.r == 0:
            self.weight = self.register_parameter("weight", None)
            self.bias = self.register_parameter("bias", None)
        else:
            self.weight = torch.nn.Parameter(torch.ones(in_dim, requires_grad=True))
            self.bias = torch.nn.Parameter(torch.zeros(in_dim, requires_grad=True))

    def forward(
        self,
        t: torch.Tensor,
    ) -> torch.Tensor:
        if self.r == 0:
            return self.activate(t)
        else:
            return self.tensor_activate(t, r=self.r)

    def activate(self, t: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError(
            f"{self.__class__.__name__} must have 'activate'!"
        )

    def tensor_activate(self, t: torch.Tensor, r: int) -> torch.Tensor:
        raise NotImplementedError(
            f"{self.__class__.__name__} must have 'tensor_activate'!"
        )

class TensorNormSilu(NormGateBase):
    """
    silu(x) = x * sigmoid(x), gate = F.sigmoid(norm)
    """
    def activate(self, t: torch.Tensor) -> torch.Tensor:
        return F.silu(t)

    def tensor_activate(self, t: torch.Tensor, r: int) -> torch.Tensor:
        B, C = t.size(0), t.size(1)
        t_flat = t.view(B, C, -1)
        norm = self.weight * t_flat.pow(2).sum(dim=-1) + self.bias
        gate = torch.sigmoid(norm)
        return expand_dims_to(gate, 2 + r) * t
    
class TensorNormTanh(NormGateBase):

    def activate(self, t: torch.Tensor) -> torch.Tensor:
        return torch.tanh(t)

    def tensor_activate(self, t: torch.Tensor, r: int) -> torch.Tensor:
        B, C = t.size(0), t.size(1)
        t_flat = t.view(B, C, -1)
        norm = self.weight * t_flat.pow(2).sum(dim=-1) + self.bias
        gate = torch.tanh(norm) / torch.where(norm == 0, torch.ones_like(norm), norm)
        return expand_dims_to(gate, 2 + r) * t

class TensorNormRelu(NormGateBase):

    def activate(self, t: torch.Tensor) -> torch.Tensor:
        return F.relu(t)
    
    def tensor_activate(self, t: torch.Tensor, r: int) -> torch.Tensor:
        B, C = t.size(0), t.size(1)
        t_flat = t.view(B, C, -1)
        norm = self.weight * t_flat.pow(2).sum(dim=-1) + self.bias
        gate = torch.heaviside(norm, torch.zeros_like(norm))
        return expand_dims_to(gate, 2 + r) * t

class TensorNormJilu(NormGateBase):
    """
    Similar to TensorSilu, but use tanh(norm) as gate so the factor could be negative
    """
    def activate(self, t: torch.Tensor) -> torch.Tensor:
        return F.silu(t)

    def tensor_activate(self, t: torch.Tensor, r: int) -> torch.Tensor:
        B, C = t.size(0), t.size(1)
        t_flat = t.view(B, C, -1)
        norm = self.weight * t_flat.pow(2).sum(dim=-1) + self.bias
        gate = F.tanh(norm)
        return expand_dims_to(gate, 2 + r) * t
    
class TensorNormIdentity(torch.nn.Module):

    def activate(self, t: torch.Tensor) -> torch.Tensor:
        return t

    def tensor_activate(self, t: torch.Tensor, r: int) -> torch.Tensor:
        return t

NormGate: Dict[str | None, type[torch.nn.Module]] = {
    "silu": TensorNormSilu,
    "jilu": TensorNormJilu,
    "tanh": TensorNormTanh,
    "relu": TensorNormRelu,
    "None": TensorNormIdentity,
    "null": TensorNormIdentity,
    None: TensorNormIdentity,
}

class GatedGate(torch.nn.Module):
    """
    shape [batch, channel] + [3,] * l
    """
    r: int
    gate: str
    act: torch.nn.Module

    def __init__(
        self,
        r: int,
        gate: str,
    ) -> None:
        super().__init__()

        self.r = r
        self.gate = gate
        self.act = ACT[gate]()

    def forward(self, gate: torch.Tensor, gated) -> torch.Tensor:
        if self.r == 0:
            return self.activate(gated)
        else:
            return self.tensor_activate(gate, gated)

    def activate(self, gated: torch.Tensor) -> torch.Tensor:
       return self.act(gated)

    def tensor_activate(self, gate: torch.Tensor, gated: int) -> torch.Tensor:
        gate = expand_dims_to(self.act(gate), self.r+2)
        return gate * gated
    




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import gc
import yaml
import logging
from typing import Dict, List
from pathlib import Path

import torch
from tqdm import tqdm
from hydra.utils import instantiate
from lightning.pytorch.utilities.rank_zero import rank_zero_only

from .element import build_element_lookup
from .read import _read
from .graph import from_atoms
from .statistics import compute_atomic_energy, _compute_statistics, Statistics
from .quantity import KeySpecification


@rank_zero_only
def create_graphs_for_main_rank(atomsList, element, for_dataset, stage):
    dataset = []
    for atoms in tqdm(atomsList, desc=f"Building graphs for {stage}"):
        dataset.append(from_atoms(element, atoms, **for_dataset))
    return dataset

@rank_zero_only
def build_atomsList(
    cfg: dict,
    target_property: List[str],
    embedding_property: List[str],
    keyspec: KeySpecification,
    num_levels: int,
):
    threeAtomsList = _read(cfg, target_property, embedding_property, keyspec)

    # ==== read atomic_numbers and atomic_energy from dataset and cfg ===
    try:
        atomsList = (
            (threeAtomsList[0])
            if threeAtomsList[1] is None
            else threeAtomsList[0] + threeAtomsList[1]
        )
        atomic_numbers_from_dataset = set(
            int(atomic_number)
            for atoms in atomsList
            for atomic_number in atoms.get_atomic_numbers()
        )
    except Exception as e:
        raise RuntimeError(f"Failed to extract atomic numbers from dataset: {e}")

    atomic_numbers = cfg['model']['config'].get("atomic_numbers", None)
    if  atomic_numbers is not None:
        atomic_numbers_from_cfg = set(atomic_numbers)
        assert atomic_numbers_from_dataset.issubset(atomic_numbers_from_cfg), (
            f"cfg.model.config.atomic_numbers must include all atomic numbers present in the dataset, "
            f"but is missing: {atomic_numbers_from_dataset - atomic_numbers_from_cfg}"
        )
        atomic_numbers_from_dataset = atomic_numbers_from_cfg

    # === multi-level atomic_energy ===
    atomic_energies= []
    if "energy" in target_property:

        num_levels = cfg['model']['config'].get("num_levels", 1)
        atomic_energies_cfg = cfg['model']['config'].get("atomic_energies", None)
        assert atomic_energies_cfg is None or isinstance(atomic_energies_cfg, List)

        if atomic_energies_cfg is not None:
            assert num_levels == len(atomic_energies_cfg)
            for v in atomic_energies_cfg:
                assert isinstance(v, Dict), "If you want to use multi-fidelity or multi-head training, "
                "you must provide each level's atomic energy or set null"
            atomic_numbers_from_energy = set(z for IAE_dict in atomic_energies_cfg for z in IAE_dict.keys())
            atomic_numbers_from_dataset = atomic_numbers_from_dataset | atomic_numbers_from_energy

        element = build_element_lookup(atomic_numbers_from_dataset)

        if atomic_energies_cfg is None:
            logging.info("Computing Isolated Atomic Energies (IAE) automatically for each level")
            atomic_energies = compute_atomic_energy(
                threeAtomsList[0], 
                element, 
                keyspec,
                num_levels
            ) 
        else:
            for idx, energy_cfg in enumerate(atomic_energies_cfg):
                atomic_energy = {int(k): float(v) for k, v in energy_cfg.items()}
                atomic_energy_keys = set(atomic_energy.keys())
                assert atomic_energy_keys.issubset(atomic_numbers_from_dataset), (
                    f"Level {idx}: Keys in atomic_energy must be subset of dataset atomic numbers. "
                    f"Unexpected: {atomic_energy_keys - atomic_numbers_from_dataset}"
                )
                for z in atomic_numbers_from_dataset:
                    if z not in atomic_energy:
                        atomic_energy[z] = 0.0
                        logging.warning(
                            f"Level {idx}: No isolated atomic energy provided for Z={z}, using 0.0 as default."
                        )
                atomic_energies.append(atomic_energy)

        logging.info("Isolated Atomic Energies per computational level:")
        for idx, energy in enumerate(atomic_energies):
            logging.info(f"  Level {idx}: {energy}")
    else:
        element = build_element_lookup(atomic_numbers_from_dataset)
    return element, threeAtomsList, atomic_energies

@rank_zero_only
def compute_statistics(
    cfg: Dict,
    target_property: List[str],
    embedding_property: List[str],
    keyspec: KeySpecification,
    num_levels: int,
    element,
    threeAtomsList,
    atomic_energies,
):
    # === compute statistics ===
    statistics_yaml = [Path('.') / f'statistics_{i}.yaml' for i in range(num_levels)]
    all_exist = all(p.exists() for p in statistics_yaml)
    if all_exist:
        with open(statistics_yaml[0], "r") as f:
            statistics_data = yaml.safe_load(f)
        if element.atomic_numbers == statistics_data.get('atomic_numbers', []):
            recompute = False
        else:
            recompute = True
    else:
        recompute = True

    if recompute:
        for_dataset = {
            "cutoff": float(cfg['model']['config'].get("cutoff", 6.0)),
            "max_neighbors": cfg['model']['config'].get("max_neighbors", None),
            "keyspec": keyspec,
            "target_property": target_property,
            "embedding_property": embedding_property,
            "universal_embedding": cfg.get("model", {})
            .get("config", {})
            .get("universal_embedding", None),
            "neighborlist_backend": cfg.get("dataset", {}).get("neighborlist_backend", "matscipy"),
        }

        dataloader_train = None
        dataset_train = create_graphs_for_main_rank(threeAtomsList[0], element, for_dataset, 'train')
        dataloader_train = instantiate(
            cfg["dataset"]["train_dataloader"],
            dataset=dataset_train
        )
        logging.info(f"Number of configs in train: {len(threeAtomsList[0])}")

        statistics = _compute_statistics(
            dataloader_train,
            # dataloader_valid,
            sorted(element.atomic_numbers),
            atomic_energies,
            target_property=target_property,
            device=cfg.get("misc", {}).get("device", "cpu"),
            num_levels=num_levels,
        )
    else:
        # To ensure complete reproducibility, since the statistics were originally 
        # computed by going through the training dataloader once, we need to go through 
        # it again here to ensure consistency.
        with torch.no_grad():
            for data in dataloader_train:
                batch = data['batch']
        statistics = []
        for yaml_file in statistics_yaml:
            with open(yaml_file, "r") as f:
                statistics_data = yaml.safe_load(f)
            statistics.append(Statistics(**statistics_data))

    del dataloader_train
    gc.collect()
    
    return statistics



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import List, Optional


import torch
from omegaconf import ListConfig


from .fn import LOSS_FN


class NormalLoss(torch.nn.Module):
    def __init__(
        self,
        loss_property: List[str],
        loss_function_name: List[str],
        loss_property_weights: Optional[List[float]],
        normalize: bool = False,
        **kwargs,
    ):
        super().__init__()
        assert isinstance(
            loss_property, (List, ListConfig)
        ), f"``cfg.loss.loss_property`` should be a list, got {type(loss_property)}"
        assert isinstance(
            loss_function_name, (List, ListConfig)
        ), f"``cfg.loss.loss_function_name`` should be a list, got {type(loss_property)}"
        if loss_property_weights is None:
            loss_property_weights = [0.0] * len(loss_property)
        assert isinstance(
            loss_property_weights, (List, ListConfig)
        ), f"``cfg.loss.loss_property_weights`` should be a list, got {type(loss_property)}"
        assert (
            len(loss_property) == len(loss_function_name) == len(loss_property_weights)
        )
        for fn in loss_function_name:
            assert (
                fn in LOSS_FN
            ), f"{fn} not in LOSS_FN, add by yourself, all available function name are {list(LOSS_FN)}"
        self.loss_property = loss_property
        self.loss_function_name = loss_function_name
        if normalize:
            normalizer = sum(loss_property_weights)
            self.loss_property_weights = [w / normalizer for w in loss_property_weights]
        else:
            self.loss_property_weights = loss_property_weights

    def forward(self, pred, label):
        total_loss = 0.0
        for i, func_name in enumerate(self.loss_function_name):
            loss = LOSS_FN[func_name](pred, label)
            total_loss += loss * self.loss_property_weights[i]
        return total_loss

    def __repr__(self):
        task_strs = [
            f"{p}: {fn} (weight={w:.4f})"
            for p, fn, w in zip(
                self.loss_property, self.loss_function_name, self.loss_property_weights
            )
        ]
        tasks_info = ", ".join(task_strs)
        return f"{self.__class__.__name__}({tasks_info})"



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import Dict, Callable, List


import torch
from torch import Tensor

LOSS_FN: Dict[str, Callable] = {}


def register_loss(func: Callable = None, *, key: str = None):
    def decorator(f: Callable):
        k = key if key is not None else f.__name__
        LOSS_FN[k] = f
        return f

    if func is None:
        return decorator
    else:
        return decorator(func)


@register_loss
def mse_energy(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = label.entropy * label.energy_weight
    return torch.mean(torch.square((label["energy"] - pred["energy"])) * total_weight)


@register_loss
def mse_energy_per_atom(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = label.entropy * label.energy_weight
    return torch.mean(
        torch.square(
            (label["energy"] - pred["energy"]) / (label.ptr[1:] - label.ptr[:-1])
        )
        * total_weight
    )

@register_loss
def mae_energy_per_atom(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = label.entropy * label.energy_weight
    return torch.mean(
        torch.abs((label["energy"] - pred["energy"]) / (label.ptr[1:] - label.ptr[:-1]))
        * total_weight
    )

@register_loss
def mse_forces(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    batch = label.batch
    total_weight = label.entropy[batch].unsqueeze(-1) * label.forces_weight[
        batch
    ].unsqueeze(-1)
    return torch.mean(torch.square(pred["forces"] - label["forces"]) * total_weight)


@register_loss
def mean_normed_forces(pred, label) -> Tensor:
    return torch.mean(
        torch.linalg.vector_norm(pred["forces"] - label["forces"], ord=2, dim=-1)
    )


@register_loss
def conditional_huber_forces(
    pred_forces: Tensor, label_forces: Tensor, huber_delta: float
) -> Tensor:
    "From MACE https://github.com/ACEsuit/mace"
    factors = huber_delta * torch.tensor(
        [1.0, 0.7, 0.4, 0.1], device=label_forces.device, dtype=label_forces.dtype
    )
    norm_forces = torch.norm(label_forces, dim=-1)
    c1 = norm_forces < 100
    c2 = (norm_forces >= 100) & (norm_forces < 200)
    c3 = (norm_forces >= 200) & (norm_forces < 300)
    c4 = ~(c1 | c2 | c3)
    se = torch.zeros_like(pred_forces)
    se[c1] = torch.nn.functional.huber_loss(
        label_forces[c1], pred_forces[c1], reduction="none", delta=factors[0]
    )
    se[c2] = torch.nn.functional.huber_loss(
        label_forces[c2], pred_forces[c2], reduction="none", delta=factors[1]
    )
    se[c3] = torch.nn.functional.huber_loss(
        label_forces[c3], pred_forces[c3], reduction="none", delta=factors[2]
    )
    se[c4] = torch.nn.functional.huber_loss(
        label_forces[c4], pred_forces[c4], reduction="none", delta=factors[3]
    )
    return torch.mean(se)


@register_loss
def mse_stress(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = label.entropy * label.stress_weight
    return torch.mean(
        torch.square(pred["stress"] - label["stress"])
        * total_weight.unsqueeze(-1).unsqueeze(-1)
    )


@register_loss
def mse_virials_per_atom(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = label.entropy * label.virials_weight
    num_atoms = (label.ptr[1:] - label.ptr[:-1]).view(-1, 1, 1)
    return torch.mean(
        (torch.square(pred["virials"] - label["virials"]) / num_atoms)
        * total_weight.unsqueeze(-1).unsqueeze(-1)
    )


@register_loss
def mse_direct_dipole(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = (label.entropy * label.direct_dipole_weight).unsqueeze(-1)
    return torch.mean(torch.square((label["direct_dipole"] - pred["direct_dipole"])) * total_weight)


@register_loss
def mse_direct_dipole_per_atom(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = (label.entropy * label.direct_dipole_weight).unsqueeze(-1)
    num_atoms = (label.ptr[1:] - label.ptr[:-1]).unsqueeze(-1)
    return torch.mean(
        torch.square((label["direct_dipole"] - pred["direct_dipole"]) / num_atoms) * total_weight
    )

@register_loss
def mse_conservative_dipole(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = (label.entropy * label.conservative_dipole_weight).unsqueeze(-1)
    return torch.mean(
        torch.square(label["conservative_dipole"] - pred["conservative_dipole"]) * total_weight
    )


@register_loss
def mse_conservative_dipole_per_atom(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = (label.entropy * label.conservative_dipole_weight).unsqueeze(-1)
    num_atoms = (label.ptr[1:] - label.ptr[:-1]).unsqueeze(-1)
    return torch.mean(
        torch.square((label["conservative_dipole"] - pred["conservative_dipole"]) / num_atoms) * total_weight
    )

@register_loss
def mse_direct_polarizability(
    pred: Dict[str, Tensor], label: Dict[str, Tensor]
) -> Tensor:
    total_weight = (
        (label.entropy * label.direct_polarizability_weight).unsqueeze(-1).unsqueeze(-1)
    )
    return torch.mean(
        torch.square(label["direct_polarizability"] - pred["direct_polarizability"])
        * total_weight
    )

@register_loss
def mse_direct_polarizability_per_atom(
    pred: Dict[str, Tensor], label: Dict[str, Tensor]
) -> Tensor:
    total_weight = (
        (label.entropy * label.direct_polarizability_weight).unsqueeze(-1).unsqueeze(-1)
    )
    num_atoms = (label.ptr[1:] - label.ptr[:-1]).unsqueeze(-1).unsqueeze(-1)
    return torch.mean(
        torch.square((label["direct_polarizability"] - pred["direct_polarizability"]) / num_atoms)
        * total_weight
    )

@register_loss
def mse_conservative_polarizability(
    pred: Dict[str, Tensor], label: Dict[str, Tensor]
) -> Tensor:
    total_weight = (
        (label.entropy * label.conservative_polarizability_weight).unsqueeze(-1).unsqueeze(-1)
    )
    return torch.mean(
        torch.square(label["conservative_polarizability"] - pred["conservative_polarizability"])
        * total_weight
    )

@register_loss
def mse_conservative_polarizability_per_atom(
    pred: Dict[str, Tensor], label: Dict[str, Tensor]
) -> Tensor:
    total_weight = (
        (label.entropy * label.conservative_polarizability_weight).unsqueeze(-1).unsqueeze(-1)
    )
    num_atoms = (label.ptr[1:] - label.ptr[:-1]).unsqueeze(-1).unsqueeze(-1)
    return torch.mean(
        torch.square((label["conservative_polarizability"] - pred["conservative_polarizability"]) / num_atoms)
        * total_weight
    )


@register_loss
def mse_born_effective_charges(
    pred: Dict[str, Tensor], label: Dict[str, Tensor]
) -> Tensor:
    batch = label.batch
    total_weight = (
        (label.entropy * label.born_effective_charges_weight)[batch]
        .unsqueeze(-1)
        .unsqueeze(-1)
    )

    return torch.mean(
        torch.square(label["born_effective_charges"] - pred["born_effective_charges"])
        * total_weight
    )


@register_loss
def mse_polarization_per_atom(
    pred: Dict[str, Tensor], label: Dict[str, Tensor]
) -> Tensor:
    ptr = label["ptr"]
    lattice = label["lattice"]
    num_atoms = (ptr[1:] - ptr[:-1]).reshape(-1, 1)
    error = pred["polarization"] - label["polarization"]
    error = torch.einsum("bi, bij -> bj", error, torch.linalg.inv(lattice))
    error = torch.remainder(error, 1.0)
    error = torch.where(error > 0.5, error - 1.0, error)
    error = torch.where(error < -0.5, error + 1.0, error)
    error = torch.einsum("bi, bij -> bj", error, lattice)
    return torch.mean(torch.square(error / num_atoms))


@register_loss
def mse_charges(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    batch = label.batch
    total_weight = (label.entropy * label.charges_weight)[batch]
    return torch.mean(torch.square(pred["charges"] - label["charges"]) * total_weight)


@register_loss
def mse_direct_forces(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    batch = label.batch
    total_weight = label.entropy[batch].unsqueeze(-1) * label.direct_forces_weight[
        batch
    ].unsqueeze(-1)
    return torch.mean(torch.square(pred["direct_forces"] - label["direct_forces"]) * total_weight)


@register_loss
def mse_direct_stress(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = label.entropy * label.direct_stress_weight
    return torch.mean(
        torch.square(pred["direct_stress"] - label["direct_stress"])
        * total_weight.unsqueeze(-1).unsqueeze(-1)
    )

@register_loss
def mse_direct_virials_per_atom(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    total_weight = label.entropy * label.direct_virials_weight
    num_atoms = (label.ptr[1:] - label.ptr[:-1]).view(-1, 1, 1)
    return torch.mean(
        torch.square((pred["direct_virials"] - label["direct_virials"]) / num_atoms)
        * total_weight.unsqueeze(-1).unsqueeze(-1)
    )

@register_loss
def mse_final_collinear_magmoms(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    key = "final_collinear_magmoms"
    batch = label.batch
    total_weight = (label.entropy * label.final_collinear_magmoms_weight)[batch]
    return torch.mean(torch.square(pred[key] - label[key]) * total_weight)

@register_loss
def mse_abs_final_collinear_magmoms(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    key = "final_collinear_magmoms"
    batch = label.batch
    total_weight = (label.entropy * label.final_collinear_magmoms_weight)[batch]
    return torch.mean(torch.square(abs(pred[key]) - abs(label[key])) * total_weight)

@register_loss
def mse_final_noncollinear_magmoms(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    key = "final_noncollinear_magmoms"
    batch = label.batch
    total_weight = label.entropy[batch].unsqueeze(-1) * label.final_collinear_magmoms_weight[
        batch
    ].unsqueeze(-1)
    return torch.mean(torch.square(pred[key] - label[key]) * total_weight)

@register_loss
def mse_magnetic_collinear_forces(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    key = "magnetic_collinear_forces"
    batch = label.batch
    total_weight = (label.entropy * label.magnetic_collinear_forces_weight)[batch]
    return torch.mean(torch.square(pred[key] - label[key]) * total_weight)

@register_loss
def mse_magnetic_noncollinear_forces(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    key = "magnetic_noncollinear_forces"
    batch = label.batch
    total_weight = label.entropy[batch].unsqueeze(-1) * label.magnetic_noncollinear_forces_weight[
        batch
    ].unsqueeze(-1)
    return torch.mean(
        torch.square(pred[key] - label[key]) * total_weight
    )

@register_loss
def mse_total_collinear_magmom(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    key = "total_collinear_magmom"
    total_weight = label.entropy * label.total_collinear_magmom_weight
    return torch.mean(torch.square((label[key] - pred[key])) * total_weight)

@register_loss
def mse_total_collinear_magmom_per_atom(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    key = "total_collinear_magmom"
    total_weight = label.entropy * label.total_collinear_magmom_weight
    return torch.mean(
        torch.square(
            (label[key] - pred[key]) / (label.ptr[1:] - label.ptr[:-1])
        )
        * total_weight
    )

@register_loss
def mse_total_noncollinear_magmom(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    key = "total_noncollinear_magmom"
    total_weight = (label.entropy * label.total_noncollinear_magmom_weight).unsqueeze(-1)
    return torch.mean(torch.square((label[key] - pred[key])) * total_weight)


@register_loss
def mse_total_noncollinear_magmom_per_atom(pred: Dict[str, Tensor], label: Dict[str, Tensor]) -> Tensor:
    key = "total_noncollinear_magmom"
    total_weight = (label.entropy * label.total_noncollinear_magmom_weight).unsqueeze(-1)
    num_atoms = (label.ptr[1:] - label.ptr[:-1]).unsqueeze(-1)
    return torch.mean(
        torch.square((label[key] - pred[key]) / num_atoms) * total_weight
    )


@register_loss
def l2ne_hessians(
        pred: Dict[str, Tensor], 
        label: Dict[str, Tensor],
    ) -> torch.Tensor:

    scale = 10.0
    true_hessian_flat = label["hessians"]
    num_atoms_per_graph = label["ptr"][1:] - label["ptr"][:-1]
    jacs_per_graph = pred["jacs_per_graph"]
    samples_per_graph = pred["samples_per_graph"]

    offset = 0
    losses = []

    for jac_pred, samples, n_g in zip(jacs_per_graph, samples_per_graph, num_atoms_per_graph):
        hess_size = n_g * 3 * n_g * 3
        hess_flat_g = true_hessian_flat[offset : offset + hess_size]
        hess_true = hess_flat_g.reshape(n_g, 3, n_g, 3)
        offset += hess_size

        atom_idx = samples[:, 0]
        xyz_idx = samples[:, 1]
        jac_true = hess_true[atom_idx, xyz_idx]
        diff = jac_pred - jac_true  # (k_g, n_g, 3)
        row_norm = torch.norm(diff, p=2, dim=-1)  # (k_g, n_g)
        loss_g = row_norm.sum(dim=1).mean(dim=0)
        if hess_true.abs().max().item() > 1e4:
            loss_g = loss_g * 1e-8
        losses.append(loss_g)

    total_loss = torch.stack(losses).sum()

    num_graphs = len(jacs_per_graph)
    total_loss = total_loss / num_graphs
    total_loss = total_loss * scale

    return total_loss



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import argparse
import torch


from tace.lightning import load_tace


def average_models(model_paths, ema=False):
    assert len(model_paths) > 0, "No model paths provided."
    model_avg = load_tace(
        model_paths[0], 
        device="cpu",
        strict=True,
        use_ema=ema,
    )
    model_avg.eval()
    with torch.no_grad():
        for p in model_avg.parameters():
            p.data.zero_()

    for path in model_paths:
        model = load_tace(
            path, 
            device="cpu",
            strict=True,
            use_ema=ema,
        )
        model.eval()
        with torch.no_grad():
            for p_avg, p in zip(model_avg.parameters(), model.parameters()):
                p_avg.data.add_(p.data)
    num_models = len(model_paths)
    with torch.no_grad():
        for p in model_avg.parameters():
            p.data.div_(num_models)
    return model_avg


def main():
    parser = argparse.ArgumentParser(description="SWA average TACE models")
    parser.add_argument(
        "-m", "--models",
        nargs="+",
        required=True,
        help="Paths to model checkpoints"
    )
    parser.add_argument(
        "-e", "--ema",
        type=int,
        choices=[0, 1],
        default=0,
        help="Whether use ema params, It is recommended not to use the ema parameter for averaging",
    )
    args = parser.parse_args()
    print(f"A total of {len(args.models)} models will be averaged")
    model_avg = average_models(args.models, args.ema)
    output_path = "average_model.pt"
    torch.save(model_avg, output_path)
    print(f"Averaged model saved to {output_path}")


if __name__ == "__main__":
    main()



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import Tuple, List


import torch
from e3nn.o3 import wigner_3j


def ICTD(
        n_total: int, 
        w: int = -1, # if not -1, return first rank = n_total, weight = w
        decomposition: bool = True, 
        dtype=None, 
        device=None
    ) -> Tuple[List[List[int]], List[torch.Tensor], List[torch.Tensor], List[torch.Tensor]]:
    n_now = 0
    j_now = 0
    path_list = []
    this_path = []
    this_pathmatrix = wigner_3j(0, 0, 0, dtype=dtype, device=device)
    pathmatrices_list = []
    cart2sph_list = []
    sph2cart_list = []
    stop_flag = {"stop": False}

    # generate paths and path matrices
    def paths_generate(n_now, j_now, this_path, this_pathmatrix, n_total):
        if stop_flag["stop"]:
            return
        if n_now <= n_total:
            if stop_flag["stop"]:
                return
            this_path.append(j_now)
            for j in [j_now + 1, j_now, j_now - 1]:
                if not (j_now == 0 and (j != 1)) and n_now + 1 <= n_total:
                    cgmatrix = wigner_3j(1, j_now, j, dtype=dtype, device=device)
                    this_pathmatrix_ = torch.einsum(
                        "abc,dce->dabe", this_pathmatrix, cgmatrix
                    )
                    this_pathmatrix_ = this_pathmatrix_.reshape(
                        cgmatrix.shape[0], -1, cgmatrix.shape[-1]
                    )
                    paths_generate(
                        n_now + 1, j, this_path.copy(), this_pathmatrix_, n_total
                    )
            if n_now == n_total:
                this_pathmatrix = this_pathmatrix.reshape(-1, this_pathmatrix.shape[-1])
                this_pathmatrix = this_pathmatrix * (
                    1.0 / (this_pathmatrix**2).sum(0)[0] ** (0.5)
                )  # normalize
                if w == -1:
                    pathmatrices_list.append(this_pathmatrix)
                    path_list.append(this_path)
                else:
                    if path_list:
                        path_list[-1] = this_path
                        pathmatrices_list[-1] = this_pathmatrix
                    else:
                        pathmatrices_list.append(this_pathmatrix)
                        path_list.append(this_path)
                    if this_path[-1] == w:
                        stop_flag["stop"] = True
                        return
        return
    paths_generate(n_now, j_now, this_path, this_pathmatrix, n_total)
    decomp_list = []
    for path_matrix in pathmatrices_list:
        if decomposition:
            decomp_list.append(path_matrix @ path_matrix.T)
        cart2sph_list.append(path_matrix)
        sph2cart_list.append(path_matrix.T)
    return path_list, decomp_list, cart2sph_list, sph2cart_list





################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import re
import lmdb
import pickle
import logging
import gc
import time
from pathlib import Path
from typing import List, Optional, Dict


from torch.utils.data import Dataset
from lightning.pytorch import LightningDataModule
from lightning.pytorch.utilities.rank_zero import rank_zero_only, rank_zero_info
import torch.distributed as dist
from hydra.utils import instantiate


from .graph import from_atoms
from .element import build_element_lookup, TorchElement
from .read import _read
from .statistics import Statistics
from .quantity import KeySpecification


class GraphDatasetLMDB(Dataset):
    """
    Lazy-loading LMDB dataset; each LMDB shard contains local keys 0..(n-1).
    Graphs accessed by global index are mapped to (shard_idx, local_idx).
    """
    def __init__(self, lmdb_paths: List[Path], in_memory: bool = False, cache_size: int = 1024):
        self.lmdb_paths = list(lmdb_paths)
        self.in_memory = in_memory
        self.cache_size = cache_size
        self.cache = {}  # (shard_idx, local_idx) -> graph
        self.envs = [None] * len(self.lmdb_paths)
        self.txns = [None] * len(self.lmdb_paths)
        self.lengths = []

        if in_memory:
            self.data_list = []
            self.length = 0
        else:
            self.length = 0
            self.lengths = [None] * len(self.lmdb_paths)

    def __len__(self):
        if self.in_memory:
            return self.length

        for i, l in enumerate(self.lengths):
            if l is None:
                env = lmdb.open(str(self.lmdb_paths[i]), readonly=True, lock=False, readahead=False, max_readers=64)
                with env.begin(write=False) as txn:
                    self.lengths[i] = txn.stat()["entries"]
                env.close()
        return sum(self.lengths)

    def __getitem__(self, idx):
        if self.in_memory:
            return self.data_list[idx]

        if not self.lengths:
            # safety: ensure lengths computed in not memory mode
            _ = len(self)

        # map global idx -> shard_idx, local_idx
        shard_idx = 0
        while idx >= self.lengths[shard_idx] and shard_idx < len(self.lengths):
            idx -= self.lengths[shard_idx]
            shard_idx += 1

        if shard_idx >= len(self.lmdb_paths):
            raise IndexError(f"Index {idx} out of range for dataset with lengths {self.lengths}")

        # lazy open env and txn
        if self.envs[shard_idx] is None:
            max_retries = 3
            for attempt in range(max_retries):
                try:
                    self.envs[shard_idx] = lmdb.open(
                        str(self.lmdb_paths[shard_idx]), readonly=True, lock=False,
                        readahead=False, max_readers=64
                    )
                    self.txns[shard_idx] = self.envs[shard_idx].begin(write=False)
                    break
                except lmdb.Error as e:
                    if attempt == max_retries - 1:
                        raise
                    time.sleep(1)

        key_tuple = (shard_idx, idx)
        if key_tuple in self.cache:
            return self.cache[key_tuple]

        key = f"{idx:09d}".encode("ascii")
        value = self.txns[shard_idx].get(key)
        if value is None:
            raise IndexError(f"Index {idx} not found in LMDB shard {self.lmdb_paths[shard_idx]} (shard_idx {shard_idx})")

        graph = pickle.loads(value)

        if len(self.cache) >= self.cache_size:
            self.cache.pop(next(iter(self.cache)))
        self.cache[key_tuple] = graph
        return graph


def create_graphs(
    atoms_list: Optional[List],
    element,
    for_dataset: dict,
    stage: str,
    shard_dirs: List[Path],
    storage_mode: str = "lmdb",
    shard_size: int = 1000,
    lmdb_wait_timeout: int = 86400,  # seconds for non-rank0 to wait for lmdb files, 1 day
    avg_graph_size_in_KB: int = 75
):
    """
    Create graphs in memory (memory mode) or write/read LMDB shards (lmdb mode).

    atoms_list may be None for ranks that don't hold the raw atoms (e.g., non-rank0 in lmdb mode).
    """
    if dist.is_initialized():
        rank = dist.get_rank()
        world_size = dist.get_world_size()
    else:
        rank = 0
        world_size = 1

    if storage_mode not in ["lmdb", "memory"]:
        raise ValueError(f"Unknown storage_mode: {storage_mode}")

    # === MEMORY MODE ===
    if storage_mode == "memory":
        if rank == 0:
            graphs = [from_atoms(element, atoms, **for_dataset) for atoms in (atoms_list or [])]
        else:
            graphs = []

        if world_size > 1 and dist.is_initialized():
            # broadcast number then chunks via broadcast_object_list in batches
            if rank == 0:
                n_obj = [len(graphs)]
            else:
                n_obj = [None]
            dist.broadcast_object_list(n_obj, src=0)
            n = int(n_obj[0])

            # initialize empty list on non-zero ranks
            if rank != 0:
                graphs = [None] * n

            # ===== broadcast in batches =====
            batch_size = 128
            for i in range(0, n, batch_size):
                j = min(n, i + batch_size)
                batch = graphs[i:j] if rank == 0 else [None] * (j - i)
                dist.broadcast_object_list(batch, src=0)
                if rank != 0:
                    graphs[i:j] = batch

        dataset = GraphDatasetLMDB(lmdb_paths=[], in_memory=True)
        dataset.data_list = graphs
        dataset.length = len(graphs)
        return dataset


    # === LMDB MODE ===
    def _collect_lmdb_paths():
        out = []
        for d in shard_dirs:
            out.extend(sorted(d.glob(f"{stage}_shard*.lmdb")))
        return out

    lmdb_paths = _collect_lmdb_paths()
    
    # if shards already exist, load them
    if lmdb_paths:
        rank_zero_info(f"Found existing LMDB files for {stage}: {len(lmdb_paths)} shards")
        if world_size > 1 and dist.is_initialized():
            dist.barrier()
        return GraphDatasetLMDB(lmdb_paths, in_memory=False)

    # Only rank 0 will create LMDB shards (if atoms_list provided)
    if rank == 0 and atoms_list:
        rank_zero_info(f"Rank 0 building and saving {len(atoms_list)} graphs for stage={stage} (stream LMDB mode)")
        buffer = []
        shard_idx = 0
        n_atoms = len(atoms_list)
        # choose directory as round-robin across provided shard_dirs for distribution
        n_dirs = max(1, len(shard_dirs))
        for d in shard_dirs:
            d.mkdir(parents=True, exist_ok=True)

        for idx, atoms in enumerate(atoms_list):
            graph = from_atoms(element, atoms, **for_dataset)
            buffer.append(graph)

            if len(buffer) >= shard_size or idx == n_atoms - 1:
                dir_path = shard_dirs[shard_idx % n_dirs]
                lmdb_path = dir_path / f"{stage}_shard{shard_idx:04d}.lmdb"
                avg_graph_size = avg_graph_size_in_KB * 1024
                map_size = int(len(buffer) * avg_graph_size * 1.1)

                max_retries = 3
                for attempt in range(max_retries):
                    try:
                        env = lmdb.open(
                            str(lmdb_path),
                            map_size=map_size,
                            subdir=True,
                            writemap=True,
                            metasync=True,
                            sync=True,
                            map_async=False,
                        )
                        with env.begin(write=True) as txn:
                            # local keys start from 0 for each shard
                            for i_local, graph_obj in enumerate(buffer):
                                key = f"{i_local:09d}".encode("ascii")
                                txn.put(key, pickle.dumps(graph_obj))
                        env.close()
                        break
                    except lmdb.Error as e:
                        if attempt == max_retries - 1:
                            # on persistent failure, clean and re-raise
                            if lmdb_path.exists():
                                import shutil
                                try:
                                    shutil.rmtree(lmdb_path)
                                except Exception:
                                    pass
                            raise
                        time.sleep(1)

                rank_zero_info(f"Saved shard {lmdb_path} with {len(buffer)} graphs")
                buffer.clear()
                shard_idx += 1
                gc.collect()

        rank_zero_info(f"All {n_atoms} graphs saved successfully for stage={stage}")
        gc.collect()

        #  other ranks wait for files to appear 
    if world_size > 1 and dist.is_initialized():
        if rank != 0:
            # poll for lmdb files with timeout to avoid indefinite hang
            waited = 0
            poll_interval = 1.0
            while waited < lmdb_wait_timeout:
                lmdb_paths = _collect_lmdb_paths()
                if lmdb_paths:
                    break
                time.sleep(poll_interval)
                waited += poll_interval
            if not lmdb_paths:
                # after timeout, still no files -> error to avoid silent hang
                raise RuntimeError(f"Timeout: no LMDB files for stage={stage} found after {lmdb_wait_timeout}s")
        # ensure all processes reach here after files exist or after rank0 created them
        dist.barrier()

    lmdb_paths = _collect_lmdb_paths()
    if not lmdb_paths:
        logging.error(f"No LMDB files found for {stage} after creation")
        raise RuntimeError(f"No LMDB files found for {stage} after creation")

    rank_zero_info(f"Loaded {len(lmdb_paths)} LMDB shards for {stage}")
    dataset = GraphDatasetLMDB(lmdb_paths, in_memory=False)
    return dataset


# === DataModule ===
class GraphDataModule(LightningDataModule):
    def __init__(
        self,
        cfg: dict,
        element: TorchElement,
        statistics: Statistics,
        target_property: List[str],
        keyspec: KeySpecification,
        embedding_property: List[str],
        num_levels: int = 1
    ):
        super().__init__()
        self.cfg = cfg
        self.keyspec = keyspec
        self.element = element
        self.num_levels = num_levels
        self.statistics = statistics
        self.target_property = target_property
        self.embedding_property = embedding_property

        self.train_dataset = None
        self.val_dataset = None
        self.test_datasets = None
        self.threeAtomsList = None

        self.storage_mode = cfg.get("dataset", {}).get("storage_mode", "memory")
        self.shard_dirs = [Path(p) for p in cfg.get("dataset", {}).get("shard_dirs", ["graphCache"])]
        self.presave_test_sets = any(
            any(subdir.is_dir() and subdir.name.startswith("test") for subdir in shard_dir.iterdir())
            for shard_dir in self.shard_dirs if shard_dir.exists()
        )
        lmdb_test_setspattern = re.compile(r"^test(\d+)_shard\d+\.lmdb$")
        test_ids = set()
        for shard_dir in self.shard_dirs:
            if not shard_dir.exists():
                continue
            for f in shard_dir.iterdir():
                if f.is_dir():
                    m = lmdb_test_setspattern.match(f.name)
                    if m:
                        test_ids.add(int(m.group(1)))
        self.num_test_sets = len(test_ids)
        self.shard_size = cfg.get("dataset", {}).get("shard_size", 1_000_000)
        self.avg_graph_size_in_KB = cfg.get("dataset", {}).get("avg_graph_size_in_KB", 200)
        self.lmdb_wait_timeout = cfg.get("dataset", {}).get("lmdb_wait_timeout", 86400)
        self.no_valid_set = cfg.get("dataset", {}).get("no_valid_set", False)
        self.neighborlist_backend = self.cfg.get("dataset", {}).get("neighborlist_backend", "matscipy")
        self._for_dataset_config = {
            "cutoff": float(cfg['model']['config'].get("cutoff", 6.0)),
            "max_neighbors": cfg['model']['config'].get("max_neighbors", None),
            "keyspec": self.keyspec,
            "target_property": self.target_property,
            "embedding_property": self.embedding_property,
            "universal_embedding": self.cfg.get("model", {}).get("config", {}).get("universal_embedding", None),
            "neighborlist_backend": self.neighborlist_backend,
        }
        logging.info(f"Neighborlist backend is {self.neighborlist_backend}")


    @rank_zero_only
    def prepare_data(self):
        if self.storage_mode == 'lmdb':
            for d in self.shard_dirs:
                d.mkdir(parents=True, exist_ok=True)
            lmdb_exists = any(list(d.rglob("*.lmdb")) for d in self.shard_dirs)
            if lmdb_exists:
                logging.info("[prepare_data] LMDB directories already exist — skipping raw data reading.")
                self.threeAtomsList = None
                return

        logging.info("[prepare_data] Reading raw atoms from source files...")
        self.threeAtomsList = _read(
            self.cfg,
            self.target_property,
            self.embedding_property,
            self.keyspec,
            in_datamodule=True,
        )

    def setup(self, stage: Optional[str] = None):
        """
        Key logic:
         - In LMDB mode: do NOT broadcast threeAtomsList. Only rank0 will have it.
           All ranks will call create_graphs(...) (with atoms_list possibly None), which
           will either find existing LMDB shards (and load) or rank0 will create them.
         - In memory mode: broadcast threeAtomsList to other ranks.
        """
        world_size = dist.get_world_size() if dist.is_initialized() else 1
        rank = dist.get_rank() if dist.is_initialized() else 0

        # === memory mode ===
        if self.storage_mode != "lmdb" and world_size > 1 and dist.is_initialized():
            if rank == 0:
                objects = [self.threeAtomsList]
            else:
                objects = [None]
            logging.info("Start broadcast of threeAtomsList (memory mode)")
            dist.broadcast_object_list(objects, src=0)
            logging.info("End broadcast of threeAtomsList (memory mode)")
            self.threeAtomsList = objects[0]
            dist.barrier()

        if stage in (None, "fit"):
            # TRAIN
            atoms_for_train = (self.threeAtomsList[0] if (self.threeAtomsList and self.threeAtomsList[0] is not None) else None)
            self.train_dataset = create_graphs(
                atoms_for_train,
                self.element,
                self._for_dataset_config,
                "train",
                self.shard_dirs,
                storage_mode=self.storage_mode,
                shard_size=self.shard_size,
                lmdb_wait_timeout=self.lmdb_wait_timeout,
                avg_graph_size_in_KB=self.avg_graph_size_in_KB,
            )

            logging.info(f"Rank {rank}: Number of configs in train: {len(self.train_dataset)}")
            if rank == 0 and self.threeAtomsList and self.threeAtomsList[0] is not None:
                self.threeAtomsList[0] = None
                gc.collect()

            # VALID
            if not self.no_valid_set:
                atoms_for_valid = (self.threeAtomsList[1] if (self.threeAtomsList and self.threeAtomsList[1] is not None) else None)
                self.val_dataset = create_graphs(
                    atoms_for_valid,
                    self.element,
                    self._for_dataset_config,
                    "valid",
                    self.shard_dirs,
                    storage_mode=self.storage_mode,
                    shard_size=self.shard_size,
                    lmdb_wait_timeout=self.lmdb_wait_timeout,
                    avg_graph_size_in_KB=self.avg_graph_size_in_KB,
                )
                logging.info(f"Rank {rank}: Number of configs in valid: {len(self.val_dataset)}")
                if rank == 0 and self.threeAtomsList and self.threeAtomsList[1] is not None:
                    self.threeAtomsList[1] = None
                    gc.collect()
            else:
                self.val_dataset = None

            # TEST
            atoms_for_test_container = (self.threeAtomsList[2] if (self.threeAtomsList and self.threeAtomsList[2] is not None) else None)
            self.test_datasets = []
            if not self.presave_test_sets:
                if atoms_for_test_container is not None:
                    for idx, test_atoms_list in enumerate(atoms_for_test_container):
                        test_dataset = create_graphs(
                            test_atoms_list,
                            self.element,
                            self._for_dataset_config,
                            f"test{idx}",
                            self.shard_dirs,
                            storage_mode=self.storage_mode,
                            shard_size=self.shard_size,
                            lmdb_wait_timeout=self.lmdb_wait_timeout,
                            avg_graph_size_in_KB=self.avg_graph_size_in_KB,
                        )
                        self.test_datasets.append(test_dataset)
                    self.presave_test_sets = True
            else:
                for idx in range(self.num_test_sets):
                    test_dataset = create_graphs(
                        None,
                        self.element,
                        self._for_dataset_config,
                        f"test{idx}",
                        self.shard_dirs,
                        storage_mode=self.storage_mode,
                        shard_size=self.shard_size,
                        lmdb_wait_timeout=self.lmdb_wait_timeout,
                        avg_graph_size_in_KB=self.avg_graph_size_in_KB,
                    )
                    self.test_datasets.append(test_dataset)
            if rank == 0 and self.threeAtomsList and self.threeAtomsList[2] is not None:
                self.threeAtomsList[2] = None
                gc.collect()

                
    def train_dataloader(self):
        return instantiate(self.cfg["dataset"]["train_dataloader"], dataset=self.train_dataset)

    def val_dataloader(self):
        if self.no_valid_set or self.val_dataset is None: 
            # if None, will warning dataloader's length is zero, just ignore it
            return instantiate(self.cfg["dataset"]["valid_dataloader"], dataset=[])
        return instantiate(self.cfg["dataset"]["valid_dataloader"], dataset=self.val_dataset)

    def test_dataloader(self):
        if self.test_datasets is None:
            return None
        return [instantiate(self.cfg["dataset"]["test_dataloader"], dataset=td) for td in self.test_datasets]


# === Datamodule Builder ===
def build_datamodule(
    cfg: Dict,
    target_property: List[str],
    embedding_property: List[str],
    keyspec: KeySpecification,
    num_levels: int,
    statistics: List[Statistics],
):
    element = build_element_lookup(statistics[0]['atomic_numbers'])
    datamodule = GraphDataModule(
        cfg,
        element,
        statistics,
        target_property,
        keyspec,
        embedding_property,
        num_levels,
    )
    return datamodule



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
import math
from pathlib import Path


from omegaconf import OmegaConf
from hydra.utils import to_absolute_path


def _resolve_path(path_str: str, expect: str = "file") -> str:
    path = Path(to_absolute_path(path_str))

    if expect == "file":
        if not path.is_file():
            raise FileNotFoundError(f"Expected file not found: {path}")
    elif expect == "dir":
        if not path.exists():
            print(f"[INFO] Directory not found, creating: {path}")
            path.mkdir(parents=True, exist_ok=True)
        elif not path.is_dir():
            raise NotADirectoryError(f"Expected directory but found file: {path}")
    else:
        raise ValueError(f"Unknown expect value: {expect} (should be file or dir)")
    return str(path)


def register_resolvers():
    OmegaConf.register_new_resolver("check_file", lambda s: _resolve_path(s, "file"))
    OmegaConf.register_new_resolver("check_dir", lambda s: _resolve_path(s, "dir"))
    OmegaConf.register_new_resolver("add", lambda a, b: float(a) + float(b))
    OmegaConf.register_new_resolver("sub", lambda a, b: float(a) - float(b))
    OmegaConf.register_new_resolver("mul", lambda a, b: float(a) * float(b))
    OmegaConf.register_new_resolver(
        "div", lambda a, b: float(a) / float(b) if float(b) != 0 else float("inf")
    )
    OmegaConf.register_new_resolver("float", lambda x: float(x))
    OmegaConf.register_new_resolver("ceil", lambda x: math.ceil(float(x)))
    OmegaConf.register_new_resolver("floor", lambda x: math.floor(float(x)))



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import Dict, List, Optional, Any, Type


import torch


from .radial import ZBLBasis
from .layers import OneHotToAtomicEnergy, ScaleShift
from .readout import build_scalar_readout, build_tensor_readout, ScalarReadOut
from .representation import TACEDescriptor
from .utils import Graph, compute_fixed_charge_dipole
from .default import (
    RADIAL_BASIS,
    ANGULAR_BASIS,
    RADIAL_MLP,
    INTER,
    PROD,
    READOUT_EMLP,
    SCALE_SHIFT,
    SHORT_RANGE,
    LONG_RANGE,
    check_model_config
)
from .basis_change import PropertyBasisChange
from ...dataset.statistics import Statistics
from ...dataset.quantity import get_target_irreps
from ...utils.torch_scatter import scatter_sum

class TACEV1(torch.nn.Module):
    def __init__(
        self,
        cutoff: float,
        statistics: List[Type[Statistics]],
        max_neighbors: Optional[int] = None,
        lmax: int | List[int] = 3,
        Lmax: int | List[int] = 2,
        bias: bool = False,
        num_layers: int = 2,
        num_levels: int = 1,
        num_channel: int = 64,
        num_channel_hidden: int = 64,
        use_multi_head: bool = False,
        use_multi_fidelity: bool = False,
        radial_basis: Dict = RADIAL_BASIS,
        radial_mlp: Dict = RADIAL_MLP,
        angular_basis: Dict = ANGULAR_BASIS,
        inter: Dict = INTER,
        prod: Dict = PROD,
        scale_shift: Dict = SCALE_SHIFT,
        short_range: Dict[str, bool] = SHORT_RANGE,
        long_range: Dict[str, bool] = LONG_RANGE,
        readout_emlp: Dict = READOUT_EMLP,
        target_property: List[str] = [],
        embedding_property: List[str] = [],
        conservation: Dict[str, Dict[str, bool | str]] = {},
        universal_embedding: Dict[str, Dict[str, str | bool | int | float]] = {},
        **kwargs,
    ):
        cfg = {
            k: v for k, v in locals().items() 
            if k != "self" 
            and not k.startswith('_')
            and not k.startswith('__')
        }
        cfg = check_model_config(cfg)
        super().__init__()
        # === init ===
        if "model_config" in kwargs:
            self.model_config = kwargs["model_config"]
        self.statistics = cfg["statistics"]
        self.max_neighbors = cfg['max_neighbors']
        self.avg_num_neighbors = cfg['avg_num_neighbors']
        self.target_property = cfg['target_property']
        self.embedding_property = cfg['embedding_property']
        self.conservation = cfg['conservation']
        self.universal_embedding = cfg['universal_embedding']
        self.target_irreps = get_target_irreps(self.target_property)
        if max(cfg['Lmax']) < max(self.target_irreps):
                raise ValueError(
                    f"cfg.model.config.Lmax {max(cfg['Lmax'])} should be greatet than"
                    f"the tensor property you want to predict."
                )
        self.register_buffer('num_layers', torch.tensor(cfg['num_layers'], dtype=torch.int64))
        self.register_buffer('atomic_numbers', torch.tensor(cfg['atomic_numbers'], dtype=torch.int64))
        self.register_buffer('cutoff', torch.tensor(cfg['cutoff'], dtype=torch.get_default_dtype()))

        for_descriptor = {
            "cutoff": cfg['cutoff'],
            "num_layers": cfg['num_layers'],
            "Lmax": cfg['Lmax'],
            "lmax": cfg['lmax'],
            "num_channel": cfg['num_channel'],
            "num_channel_hidden": cfg['num_channel_hidden'],
            "radial_basis": cfg['radial_basis'],
            "radial_mlp": cfg['radial_mlp'],
            "angular_basis": cfg['angular_basis'],
            "inter": cfg['inter'],
            "prod": cfg['prod'],
            "bias": cfg['bias'],
            "target_irreps": self.target_irreps,
            "universal_embedding": cfg['universal_embedding'],
            "atomic_numbers": cfg['atomic_numbers'],
            "avg_num_neighbors": cfg['avg_num_neighbors'],
        }
        self.descriptor = TACEDescriptor(**for_descriptor)

        # === Readout ===
        self.use_all_layer = cfg['readout_emlp'].get('use_all_layer', True)
        self.use_multi_head = cfg['use_multi_head']
        for_scalar_readout = {
            'in_dim': cfg['num_channel'],
            'hidden_dim': cfg['readout_emlp'].get('hidden_dim', [16]),
            'act': cfg['readout_emlp'].get('act', "silu"),
            'bias': cfg['readout_emlp'].get('bias', False),
            'num_levels': cfg.get('num_levels', 1),
            'use_multi_head': cfg.get('use_multi_head', False),
            'num_layers': cfg['num_layers'],
            'use_all_layer': cfg['readout_emlp'].get('use_all_layer', True)
        }

        for_tensor_readout = {
            'in_dim': cfg['num_channel'],
            'hidden_dim': cfg['readout_emlp'].get('hidden_dim', [16]),
            'gate': cfg['readout_emlp'].get('gate', "silu"),
            'num_levels': cfg.get('num_levels', 1),
            'use_multi_head': cfg.get('use_multi_head', False),
            'num_layers': cfg['num_layers'],
            'use_all_layer': cfg['readout_emlp'].get('use_all_layer', True)
        }

        # === Energy ===
        if "energy" in self.target_property:
            self.energy_readouts = build_scalar_readout(**for_scalar_readout)
            self.atomic_energy_layer = OneHotToAtomicEnergy(cfg['atomic_energies'])
            self.scale_shift = ScaleShift.build_from_config(cfg['statistics'], cfg['scale_shift'])

        # === Direct Dipolet === 
        if "direct_dipole" in self.target_property:
            self.dipole_readouts = build_tensor_readout(l=1, **for_tensor_readout)
            
        # === Direct Forces Readout ===
        if "direct_forces" in self.target_property:
            self.direct_forces_readouts = build_tensor_readout(l=1, **for_tensor_readout)
       
        # === Direct Polarizability ===
        if "direct_polarizability" in self.target_property:
            self.direct_polarizability_readout0s = build_scalar_readout(**for_scalar_readout)
            self.direct_polarizability_readout2s = build_tensor_readout(l=2, **for_tensor_readout)
            self.direct_polarizability_basis_change = PropertyBasisChange["direct_polarizability"]() 

        # === Direct Virials ===
        if 'direct_virials' in self.target_property or 'direct_stress' in self.target_property:
            self.direct_virials_readout0s = build_scalar_readout(**for_scalar_readout)
            self.direct_virials_readout2s = build_tensor_readout(l=2, **for_tensor_readout)
            self.direct_virials_basis_change = PropertyBasisChange["direct_virials"]() 

        # === Final Collinear Magmoms ===
        if 'final_collinear_magmoms' in self.target_property:
            self.final_collinear_magmoms_readouts = build_scalar_readout(**for_scalar_readout)

        # === Final Noncollinear Magmoms === not conform to time reversal ...
        if 'final_noncollinear_magmoms' in self.target_property:
            self.final_noncollinear_magmoms_readouts = build_tensor_readout(l=1, **for_tensor_readout)

        # === Charges ===
        if "charges" in self.target_property:
            self.predict_charges_method = self.conservation.get('charges', {}).get('method', 'lagrangian')
            if self.predict_charges_method == 'lagrangian':
                self.chi_readouts = build_scalar_readout(**for_scalar_readout)
                self.eta_readouts = build_scalar_readout(**for_scalar_readout)
            elif self.predict_charges_method == 'uniform_distribution':
                self.charges_readouts = build_scalar_readout(**for_scalar_readout)
            else:
                raise ValueError(
                    f"Unknown predict_charges_method: {self.predict_charges_method}. "
                    "Supported methods are ['lagrangian', 'uniform_distribution']."
                )
            
        # === Direct Hessians ===
        if 'direct_hessians' in self.target_property:
            self.direct_hessians_readout0s = build_scalar_readout(**for_scalar_readout)
            self.direct_hessians_readout2s = build_tensor_readout(l=2, **for_tensor_readout)
            self.direct_hessians_basis_change = PropertyBasisChange["direct_hessians"]() 

        # === Universal Invariant Embedding, not allow multi-head ===
        if cfg['readout_emlp'].get('enable_uie_readout', False) \
        and self.universal_embedding['invariant_embedding_property']:
            self.uie_readout = ScalarReadOut(
                cfg['num_channel'],
                1,
                hidden_dim=[],
                act=None,
                bias=False,
                forward_weight_init=True,
            )

       # === Short range, Long range ===
        short_range_ = cfg.get('short_range', {})
        long_range_ = cfg.get('long_range', {})
        if short_range_.get('zbl', {}).get('enable', True) and 'energy' in self.target_property:
            self.zbl = ZBLBasis(
                cfg['radial_basis']["polynomial_cutoff"],
                short_range_.get('zbl', {}).get('trainable', False),
            )
        if long_range_.get('les', {}).get('enable', False):
            try:
                from les import Les
            except ImportError as e:
                raise ImportError(
                    "Can not import les. Please install the library from https://github.com/ChengUCB/les."
                ) from e
            les_arguments = long_range_.get('les', {}).get('les_arguments', None)
            if les_arguments is None:
                les_arguments = {"use_atomwise": False}
            self.compute_bec = les_arguments.get("compute_bec", False)
            self.bec_output_index = les_arguments.get("bec_output_index", None)
            self.les = Les(les_arguments=les_arguments)
            self.les_readouts = build_scalar_readout(**for_scalar_readout)

    def readout_fn(
        self,
        data: Dict[str, torch.Tensor],
        graph: Graph,
        from_representation: Dict[str, Optional[torch.Tensor]]
    ) -> Dict[str, Optional[torch.Tensor]]:

        batch = data["batch"]
        descriptors = from_representation['descriptors']

        nlocal, _ = graph.lmp_natoms
        num_graphs = graph.num_graphs
        node_level = graph.node_level
        num_atoms_arange = graph.num_atoms_arange
        dtype = data['node_attrs'].dtype
        device = data['node_attrs'].device

        # === Energy ===
        E = None
        e_node = None
        if "energy" in self.target_property:
            e_base_node = self.atomic_energy_layer(data['node_attrs'])[num_atoms_arange, node_level]
            e_base_graph = scatter_sum(e_base_node, batch, dim=-1, dim_size=num_graphs)
            e_list = []
            for ii, energy_readout in enumerate(self.energy_readouts):
                if not self.use_all_layer:
                    ii = -1
                e_list.append(energy_readout(descriptors[ii][0], node_level)[num_atoms_arange, node_level])   
            e_node = torch.sum(torch.stack(e_list, dim=0), dim=0)
            # === ZBL === 
            if hasattr(self, "zbl"):
                e_zbl_node = self.zbl(
                    graph.edge_length, 
                    data['node_attrs'], 
                    data["edge_index"],
                    self.atomic_numbers
                )[num_atoms_arange]
                e_node = e_node + e_zbl_node
            # === scale and shift ===
            e_node = self.scale_shift(
                e_node, 
                data['node_attrs'][num_atoms_arange], 
                data['ptr'], 
                data['edge_index'], 
                data['batch'],
                node_level,
            )    
            # === uie === not support lammps now
            if hasattr(self, "uie_readout"):
                e_uie_node = self.uie_readout(from_representation['uie_feats'])
                e_node = e_node + e_uie_node.squeeze(-1) 
            e_graph = scatter_sum(e_node, batch, dim=-1, dim_size=num_graphs)
            e_node = e_base_node + e_node
            E = e_base_graph + e_graph

        # === Direct Forces ===
        D_F = None
        if 'direct_forces' in self.target_property:
            d_f_list = []
            for ii, direct_forces_readout in enumerate(self.direct_forces_readouts):
                if not self.use_all_layer:
                    ii = -1
                d_f_list.append(
                    direct_forces_readout(
                        descriptors[ii][1],
                        node_level,
                    )[num_atoms_arange, node_level, :]
                )
            D_F = torch.sum(torch.stack(d_f_list, dim=-1), dim=-1)

        # === Direct Dipole ===
        D = None
        if 'direct_dipole' in self.target_property:
            d_base = compute_fixed_charge_dipole(
                charges=data["charges"],
                positions=data["positions"],
                batch=data["batch"],
                num_graphs=num_graphs,
            )
            d_list = []
            for ii, dipole_readout in enumerate(self.dipole_readouts):
                if not self.use_all_layer:
                    ii = -1
                d_list.append(
                    dipole_readout(
                        descriptors[ii][1],
                        node_level,
                    )[num_atoms_arange, node_level, :]
                )
            d_node = torch.sum(torch.stack(d_list, dim=-1), dim=-1)
            d_graph = scatter_sum(d_node, batch, dim=0, dim_size=num_graphs)
            D = d_base + d_graph

        # === Direct Polarizability ===
        ALPHA = None
        if 'direct_polarizability' in self.target_property:
            alpha0_list = []; alpha2_list = []
            for ii, (polarizability_readout0, polarizability_readout2) in enumerate(
                zip(self.direct_polarizability_readout0s, self.direct_polarizability_readout2s)
            ):
                if not self.use_all_layer:
                    ii = -1
                alpha0_list.append(
                    polarizability_readout0(
                        descriptors[ii][0],
                    )[num_atoms_arange, node_level]
                )
                alpha2_list.append(
                    polarizability_readout2(
                        descriptors[ii][2],
                        node_level,
                    )[num_atoms_arange, node_level, :, :]
                )
            alpha0_node = torch.sum(torch.stack(alpha0_list, dim=-1), dim=-1)
            alpha2_node = torch.sum(torch.stack(alpha2_list, dim=-1), dim=-1)
            alpha0_graph = scatter_sum(alpha0_node, batch, dim=0, dim_size=num_graphs)
            alpha2_graph = scatter_sum(alpha2_node, batch, dim=0, dim_size=num_graphs)
            ALPHA = self.direct_polarizability_basis_change(alpha0_graph, alpha2_graph)

        # === Direct Virials and Stress ===
        D_V = None
        D_S = None
        if 'direct_virials' in self.target_property or 'direct_stress' in self.target_property:
            d_v0_list = []; d_v2_list = []
            for ii, (direct_virials_readout0, direct_virials_readout2) in enumerate(
                zip(self.direct_virials_readout0s, self.direct_virials_readout2s)
            ):
                if not self.use_all_layer:
                    ii = -1
                d_v0_list.append(
                    direct_virials_readout0(
                        descriptors[ii][0],
                    )[num_atoms_arange, node_level]
                )
                d_v2_list.append(
                    direct_virials_readout2(
                        descriptors[ii][2],
                    )[num_atoms_arange, node_level, :, :]
                )
            d_v0_node = torch.sum(torch.stack(d_v0_list, dim=-1), dim=-1)
            d_v2_node = torch.sum(torch.stack(d_v2_list, dim=-1), dim=-1)
            d_v0_graph = scatter_sum(d_v0_node, batch, dim=0, dim_size=num_graphs)
            d_v2_graph = scatter_sum(d_v2_node,batch, dim=0, dim_size=num_graphs)
            D_V = self.direct_virials_basis_change(d_v0_graph, d_v2_graph)
            VOLUME = torch.linalg.det(data["lattice"]).abs().unsqueeze(-1)
            D_S = -D_V / VOLUME.view(-1, 1, 1)
            D_S = torch.where(torch.abs(D_S) < 1e10, D_S, torch.zeros_like(D_S))

         # === Charges === 
        CHARGES = None
        if 'charges' in self.target_property:
            if self.predict_charges_method == 'lagrangian':
                chi_list = []; eta_list = []
                for ii, (chi_readout, eta_readout) in enumerate(
                    zip(self.chi_readouts, self.eta_readouts)
                ):
                    if not self.use_all_layer:
                        ii = -1
                    chi_list.append(chi_readout(descriptors[ii][0])[num_atoms_arange, node_level])
                    eta_list.append(eta_readout(descriptors[ii][0])[num_atoms_arange, node_level])
                chi_node = torch.sum(torch.stack(chi_list, dim=-1), dim=-1)
                eta_node = torch.sum(torch.stack(eta_node, dim=-1), dim=-1)
                eta_node = torch.hypot(eta_node, torch.tensor(1e-6, device=device, dtype=dtype))
                eta_node = torch.reciprocal(eta_node)
                lambda_graph = (
                    data["total_charge"]
                    + scatter_sum(
                        chi_node * eta_node, batch, dim=-1, dim_size=num_graphs
                    )
                ) / scatter_sum(eta_node, batch, dim=-1, dim_size=num_graphs)
                lambda_node = lambda_graph[batch]
                CHARGES = lambda_node * (eta_node) - (chi_node * eta_node)
            elif self.predict_charges_method == 'uniform_distribution':
                c_list = []
                for ii, charges_readout in enumerate(self.charges_readouts):
                    if not self.use_all_layer:
                        ii = -1
                    c_list.append(charges_readout(descriptors[ii][0])[num_atoms_arange, node_level])
                c_node = torch.sum(torch.stack(c_list, dim=-1), dim=-1)
                c_graph = scatter_sum(c_node, batch, dim=-1, dim_size=num_graphs)
                c_delta_node = (c_graph - data["total_charge"]) / (data["ptr"][1:] - data["ptr"][:-1])
                CHARGES = c_node + c_delta_node[batch]

        # === Direct Polarizability ===
        D_HESSIANS = None
        if 'direct_hessians' in self.target_property:
            d_hessians0_list = []; d_hessians2_list = []
            for ii, (direct_hessians_readout0, direct_hessians_readout2) in enumerate(
                zip(self.direct_hessians_readout0s, self.direct_hessians_readout2s)
            ):
                if not self.use_all_layer:
                    ii = -1
                d_hessians0_list.append(
                    direct_hessians_readout0(
                        descriptors[ii][0],
                    )[num_atoms_arange, node_level]
                )
                d_hessians2_list.append(
                    direct_hessians_readout2(
                        descriptors[ii][2],
                        node_level,
                    )[num_atoms_arange, node_level, :, :]
                )
            d_hessians0_node = torch.sum(torch.stack(d_hessians0_list, dim=-1), dim=-1)
            d_hessians2_node = torch.sum(torch.stack(d_hessians2_list, dim=-1), dim=-1)
            d_hessians_node = self.direct_hessians_basis_change(d_hessians0_node, d_hessians2_node)

            d_hessians_list = []
            for idx in range(num_graphs):
                start = data["ptr"][idx]
                end = data["ptr"][idx+1]
                this_d_hessians_node = d_hessians_node[start:end, :, :]
                d_hessians_list.append(
                    (
                        0.5 * (this_d_hessians_node[:, None, :, :] + this_d_hessians_node[None, :, :, :] )
                    ).reshape(-1, 3, 3)
                ) 
            D_HESSIANS = torch.cat(d_hessians_list, dim=0)
            
        if hasattr(self, 'les'):
            les_lq_list = []
            for ii, les_readout in enumerate(self.les_readouts):
                if not self.use_all_layer:
                    ii = -1
                les_lq_list.append(les_readout(descriptors[ii][0])[num_atoms_arange, node_level])
            LES_LQ = torch.sum(torch.stack(les_lq_list, dim=0), dim=0) 
            les_results = self.les(
                latent_charges=LES_LQ,
                positions=graph.positions,
                cell=graph.lattice.view(-1, 3, 3),
                batch=batch,
                compute_energy=True,
                compute_bec=self.compute_bec,
                bec_output_index=self.bec_output_index,
            )
            LES_E = les_results["E_lr"]
            if LES_E is None:
                LES_E = torch.zeros_like(E)
            E += LES_E
            LES_BEC = les_results["BEC"]
        else:
            LES_E = None
            LES_LQ = None
            LES_BEC = None

        scalar_descriptor = None
        if 0 in self.target_irreps:
            scalar_descriptor_list = []
            for descriptor in descriptors:
                scalar_descriptor_list.append(descriptor[0])
            scalar_descriptor = torch.cat(scalar_descriptor_list, dim=-1)

        # === Final Noncollinear Magmoms ===
        F_NC_MAG = None
        if 'final_noncollinear_magmoms' in self.target_property:
            f_nc_m_list = []
            for ii, final_noncollinear_magmoms_readout in enumerate(self.final_noncollinear_magmoms_readouts):
                if not self.use_all_layer:
                    ii = -1
                f_nc_m_list.append(
                    final_noncollinear_magmoms_readout(
                        descriptors[ii][1],
                        node_level,
                    )[num_atoms_arange, node_level, :]
                )
            F_NC_MAG = torch.sum(torch.stack(f_nc_m_list, dim=-1), dim=-1)

        # === Final Collinear Magmoms ===
        F_C_MAG = None
        if 'final_collinear_magmoms' in self.target_property:
            f_c_m_list = []
            for ii, final_collinear_magmoms_readout in enumerate(self.final_collinear_magmoms_readouts):
                if not self.use_all_layer:
                    ii = -1
                f_c_m_list.append(final_collinear_magmoms_readout(descriptors[ii][0])[num_atoms_arange, node_level])
            F_C_MAG = torch.sum(torch.stack(f_c_m_list, dim=0), dim=0)
            # print('initial', data['initial_collinear_magmoms'][:5])
            # print('final', data['final_collinear_magmoms'][:5])
            # print('predict', F_C_MAG[:5])
            
        return {
            "energy": E,
            "node_energy": e_node, # TODO, BUG check the inside of les
            "direct_dipole": D,
            "direct_polarizability": ALPHA,
            "direct_forces": D_F,
            "direct_virials": D_V,
            "direct_stress": D_S,
            "direct_hessians": D_HESSIANS,
            "charges": CHARGES,
            "les_energy": LES_E,
            "les_latent_charges": LES_LQ,
            "les_born_effective_charges": LES_BEC,
            "final_collinear_magmoms": F_C_MAG,
            "final_noncollinear_magmoms": F_NC_MAG,
            "scalar_descriptor": scalar_descriptor,
        }
    
    def forward(self, data: Dict[str, torch.Tensor], graph: Graph) -> Dict[str, Any]:
        outs = self.descriptor(data, graph)
        return self.readout_fn(data, graph, outs)
    
    def _reset_target_property(self, target_property: List[str]):
        self.target_property = target_property




###########################################################################################
# Authors: Zemin Xu
# License: Academic / Non-Commercial Use Only
###########################################################################################

import os
import glob
import importlib

__all__ = []
py_files = glob.glob(os.path.join(os.path.dirname(__file__), "*.py"))
py_files = [f for f in py_files if not f.endswith("__init__.py")]

for p in py_files:
    module_name = os.path.splitext(os.path.basename(p))[0]
    module = importlib.import_module(f".{module_name}", package=__name__)
    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        if isinstance(attr, type):
            globals()[attr_name] = attr
            __all__.append(attr_name)



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################


import logging
from pathlib import Path


import hydra
from omegaconf import DictConfig


from .train import build

@hydra.main(version_base="1.3", config_path=str(Path.cwd()), config_name="tace")
def main(cfg: DictConfig):
    _, _, _, _, datamodule = build(cfg)
    datamodule.prepare_data()
    datamodule.setup('fit')
    logging.info(f"Finished building lmdb dataset")

if __name__ == "__main__":
    main()





################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
# To avoid potential conflicts between torch and torch_scatter,
# I have directly copied the scatter operation code
# from https://github.com/rusty1s/pytorch_scatter.
"""
Copyright (c) 2020 Matthias Fey <matthias.fey@tu-dortmund.de>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from typing import Optional, Tuple

import torch


def broadcast(src: torch.Tensor, other: torch.Tensor, dim: int):
    if dim < 0:
        dim = other.dim() + dim
    if src.dim() == 1:
        for _ in range(0, dim):
            src = src.unsqueeze(0)
    for _ in range(src.dim(), other.dim()):
        src = src.unsqueeze(-1)
    src = src.expand(other.size())
    return src


def scatter_sum(src: torch.Tensor,
                index: torch.Tensor,
                dim: int = -1,
                out: Optional[torch.Tensor] = None,
                dim_size: Optional[int] = None) -> torch.Tensor:
    index = broadcast(index, src, dim)
    if out is None:
        size = list(src.size())
        if dim_size is not None:
            size[dim] = dim_size
        elif index.numel() == 0:
            size[dim] = 0
        else:
            size[dim] = int(index.max()) + 1
        out = torch.zeros(size, dtype=src.dtype, device=src.device)
        return out.scatter_add_(dim, index, src)
    else:
        return out.scatter_add_(dim, index, src)


def scatter_add(src: torch.Tensor,
                index: torch.Tensor,
                dim: int = -1,
                out: Optional[torch.Tensor] = None,
                dim_size: Optional[int] = None) -> torch.Tensor:
    return scatter_sum(src, index, dim, out, dim_size)


def scatter_mul(src: torch.Tensor,
                index: torch.Tensor,
                dim: int = -1,
                out: Optional[torch.Tensor] = None,
                dim_size: Optional[int] = None) -> torch.Tensor:
    return torch.ops.torch_scatter.scatter_mul(src, index, dim, out, dim_size)


def scatter_mean(src: torch.Tensor,
                 index: torch.Tensor,
                 dim: int = -1,
                 out: Optional[torch.Tensor] = None,
                 dim_size: Optional[int] = None) -> torch.Tensor:
    out = scatter_sum(src, index, dim, out, dim_size)
    dim_size = out.size(dim)

    index_dim = dim
    if index_dim < 0:
        index_dim = index_dim + src.dim()
    if index.dim() <= index_dim:
        index_dim = index.dim() - 1

    ones = torch.ones(index.size(), dtype=src.dtype, device=src.device)
    count = scatter_sum(ones, index, index_dim, None, dim_size)
    count[count < 1] = 1
    count = broadcast(count, out, dim)
    if out.is_floating_point():
        out.true_divide_(count)
    else:
        out.div_(count, rounding_mode='floor')
    return out


def scatter_min(
        src: torch.Tensor,
        index: torch.Tensor,
        dim: int = -1,
        out: Optional[torch.Tensor] = None,
        dim_size: Optional[int] = None) -> Tuple[torch.Tensor, torch.Tensor]:
    return torch.ops.torch_scatter.scatter_min(src, index, dim, out, dim_size)


def scatter_max(
        src: torch.Tensor,
        index: torch.Tensor,
        dim: int = -1,
        out: Optional[torch.Tensor] = None,
        dim_size: Optional[int] = None) -> Tuple[torch.Tensor, torch.Tensor]:
    return torch.ops.torch_scatter.scatter_max(src, index, dim, out, dim_size)


def scatter(src: torch.Tensor,
            index: torch.Tensor,
            dim: int = -1,
            out: Optional[torch.Tensor] = None,
            dim_size: Optional[int] = None,
            reduce: str = "sum") -> torch.Tensor:
    r"""
    |

    .. image:: https://raw.githubusercontent.com/rusty1s/pytorch_scatter/
            master/docs/source/_figures/add.svg?sanitize=true
        :align: center
        :width: 400px

    |

    Reduces all values from the :attr:`src` tensor into :attr:`out` at the
    indices specified in the :attr:`index` tensor along a given axis
    :attr:`dim`.
    For each value in :attr:`src`, its output index is specified by its index
    in :attr:`src` for dimensions outside of :attr:`dim` and by the
    corresponding value in :attr:`index` for dimension :attr:`dim`.
    The applied reduction is defined via the :attr:`reduce` argument.

    Formally, if :attr:`src` and :attr:`index` are :math:`n`-dimensional
    tensors with size :math:`(x_0, ..., x_{i-1}, x_i, x_{i+1}, ..., x_{n-1})`
    and :attr:`dim` = `i`, then :attr:`out` must be an :math:`n`-dimensional
    tensor with size :math:`(x_0, ..., x_{i-1}, y, x_{i+1}, ..., x_{n-1})`.
    Moreover, the values of :attr:`index` must be between :math:`0` and
    :math:`y - 1`, although no specific ordering of indices is required.
    The :attr:`index` tensor supports broadcasting in case its dimensions do
    not match with :attr:`src`.

    For one-dimensional tensors with :obj:`reduce="sum"`, the operation
    computes

    .. math::
        \mathrm{out}_i = \mathrm{out}_i + \sum_j~\mathrm{src}_j

    where :math:`\sum_j` is over :math:`j` such that
    :math:`\mathrm{index}_j = i`.

    .. note::

        This operation is implemented via atomic operations on the GPU and is
        therefore **non-deterministic** since the order of parallel operations
        to the same value is undetermined.
        For floating-point variables, this results in a source of variance in
        the result.

    :param src: The source tensor.
    :param index: The indices of elements to scatter.
    :param dim: The axis along which to index. (default: :obj:`-1`)
    :param out: The destination tensor.
    :param dim_size: If :attr:`out` is not given, automatically create output
        with size :attr:`dim_size` at dimension :attr:`dim`.
        If :attr:`dim_size` is not given, a minimal sized output tensor
        according to :obj:`index.max() + 1` is returned.
    :param reduce: The reduce operation (:obj:`"sum"`, :obj:`"mul"`,
        :obj:`"mean"`, :obj:`"min"` or :obj:`"max"`). (default: :obj:`"sum"`)

    :rtype: :class:`Tensor`

    .. code-block:: python

        from torch_scatter import scatter

        src = torch.randn(10, 6, 64)
        index = torch.tensor([0, 1, 0, 1, 2, 1])

        # Broadcasting in the first and last dim.
        out = scatter(src, index, dim=1, reduce="sum")

        print(out.size())

    .. code-block::

        torch.Size([10, 3, 64])
    """
    if reduce == 'sum' or reduce == 'add':
        return scatter_sum(src, index, dim, out, dim_size)
    if reduce == 'mul':
        return scatter_mul(src, index, dim, out, dim_size)
    elif reduce == 'mean':
        return scatter_mean(src, index, dim, out, dim_size)
    elif reduce == 'min':
        return scatter_min(src, index, dim, out, dim_size)[0]
    elif reduce == 'max':
        return scatter_max(src, index, dim, out, dim_size)[0]
    else:
        raise ValueError


import torch
from cartnn import o3

torch.set_printoptions(precision=1, sci_mode=False)
torch.manual_seed(3)
dtype = torch.set_default_dtype(torch.float64)


L = 3
T = torch.randn(3**L)

To_list = []
P1S, D1S, C1S, S1S = o3.ICTD(L)
for P1, D1, C1 in zip(P1S, D1S, C1S):
    To = T @ D1
    To_list.append(To)


G = torch.zeros(len(To_list), len(To_list))
for i in range(len(To_list)):
    for j in range(len(To_list)):
        print(i, j)
        G[i, j] = torch.dot(To_list[i], To_list[j])

print(G)




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import sys
import logging
from datetime import datetime

import torch
import lightning as L
from lightning.pytorch.utilities.rank_zero import rank_zero_only


from ..__init__ import __version__


LOG_LEVELS = {
    10: logging.DEBUG,
    20: logging.INFO,
    30: logging.WARNING,
    40: logging.ERROR,
    50: logging.CRITICAL,
}


def set_logger(_level: int = 20, _rank_zero_only: bool = True) -> None:
    assert _level in [10, 20, 30, 40, 50]
    root_logger = logging.getLogger()
    if root_logger.handlers:
        formatter = logging.Formatter("[%(levelname)s] %(message)s")
        for handler in root_logger.handlers:
            handler.setFormatter(formatter)
    else:
        logging.basicConfig(
            level=LOG_LEVELS[_level],
            format="[%(levelname)s] %(message)s",
        )
    root_logger.setLevel(LOG_LEVELS[_level])
    if _rank_zero_only:
        logging.debug = rank_zero_only(logging.debug)
        logging.info = rank_zero_only(logging.info)
        logging.warning = rank_zero_only(logging.warning)
        logging.error = rank_zero_only(logging.error)
        logging.critical = rank_zero_only(logging.critical)
    logging.info("Training started at %s", datetime.now().strftime("%Y-%m-%d %H:%M"))
    logging.info(
        f"\nTACE version: {__version__}\n"
        f"Torch version: {torch.__version__}\n"
        f"Lightning version: {L.__version__}\n"
        f"Python version: {sys.version}\n"
        r"""
        ,----,                                                                                             
      ,/   .`|                                                                                             
    ,`   .'  :                                                           ,---,         ,----..      ,---,. 
  ;    ;     /                                                          '  .' \       /   /   \   ,'  .' | 
.'___,/    ,'              ,---,              ,---.    __  ,-.         /  ;    '.    |   :     :,---.'   | 
|    :     |           ,-+-. /  | .--.--.    '   ,'\ ,' ,'/ /|        :  :       \   .   |  ;. /|   |   .' 
;    |.';  ;   ,---.  ,--.'|'   |/  /    '  /   /   |'  | |' |        :  |   /\   \  .   ; /--` :   :  |-, 
`----'  |  |  /     \|   |  ,"' |  :  /`./ .   ; ,. :|  |   ,'        |  :  ' ;.   : ;   | ;    :   |  ;/| 
    '   :  ; /    /  |   | /  | |  :  ;_   '   | |: :'  :  /          |  |  ;/  \   \|   : |    |   :   .' 
    |   |  '.    ' / |   | |  | |\  \    `.'   | .; :|  | '           '  :  | \  \ ,'.   | '___ |   |  |-, 
    '   :  |'   ;   /|   | |  |/  `----.   \   :    |;  : |           |  |  '  '--'  '   ; : .'|'   :  ;/| 
    ;   |.' '   |  / |   | |--'  /  /`--'  /\   \  / |  , ;           |  :  :        '   | '/  :|   |    \ 
    '---'   |   :    |   |/     '--'.     /  `----'   ---'            |  | ,'        |   :    / |   :   .' 
             \   \  /'---'        `--'---'                            `--''           \   \ .'  |   | ,'   
              `----'                                                                   `---`    `----'     
                                                                                                                       
        """
    )




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import argparse
import torch

from tace.lightning import load_tace


def main():
    parser = argparse.ArgumentParser(
        description="Load a TACE model and optionally override target properties "
                    "at inference time."
    )

    parser.add_argument(
        "-m", "--model",
        required=True,
        type=str,
        help="Path to the model checkpoint."
    )

    parser.add_argument(
        "-t", "--target_property",
        nargs="+",
        required=False,
        default=None,
        type=str,
        help=(
            "Override the model target properties at inference time. "
            "This allows requesting additional outputs that were not explicitly "
            "used as training targets. For example, one may predict hessians even "
            "if the model was trained only on forces, or enable predictions of "
            "atomic_stresses, or properties under external fields...... "
            "If not provided, the original target_property stored in the checkpoint "
            "will be used."
        )
    )
    parser.add_argument(
        "-z", "--atomic_numbers",
        nargs="+",
        required=False,
        default=None,
        type=int,
        help=(
            "List of atomic numbers. "
            "When provided, the loaded model will be modified to support only "
            "the specified elements. This is useful for exporting a smaller, "
            "element-specific model for deployment or inference on systems "
            "containing a known subset of elements. "
        )
    )

    args = parser.parse_args()
    model = load_tace(
        args.model,
        device="cpu",
        strict=True,
        use_ema=True,
    )
    if args.target_property is not None:
        print("Override target_property with:")
        model.reset_target_property(args.target_property)
        for p in model.target_property:
            print(f"  - {p}")
    else:
        print("Using target_property stored in the checkpoint:")
        for p in model.target_property:
            print(f"  - {p}")

    if args.atomic_numbers is not None:
        raise NotImplementedError("Specifying atomic numbers is not supported yet.")


if __name__ == "__main__":
    main()



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import shutil
from pathlib import Path

MAX_SIZE_BYTES = 1 * 1024**3  # 1 GB


def get_size(path: Path) -> int:
    """
    Return size of file or directory in bytes.
    - For files: directly use stat().st_size
    - For directories: sum sizes of all contained files
    """
    if path.is_file():
        return path.stat().st_size
    elif path.is_dir():
        return sum(f.stat().st_size for f in path.rglob("*") if f.is_file())
    return 0


def all_under_limit(paths) -> bool:
    """
    Check if all given paths are under MAX_SIZE_BYTES.
    Returns True if all pass the size limit, otherwise False.
    """
    for path in paths:
        if not path.exists():
            continue
        size = get_size(path)
        if size > MAX_SIZE_BYTES:
            size_gb = size / 1024**3
            print(f"⛔ Skipped: {path} size {size_gb:.2f} GB exceeds 5GB limit.")
            return False
    return True


def delete_path(path: Path):
    """
    Delete a file or directory without checking size.
    """
    if path.is_dir():
        shutil.rmtree(path)
    elif path.is_file():
        path.unlink()


def main():
    targets = [
        "train.index",
        "valid.index",
        "wandb_logs",
        "lightning_logs",
        "outputs",
        "checkpoints",
        "_tace.yaml",
        "out.txt",
        "finetune_config.yaml"
        # "graphCache",
    ]
    current_dir = Path.cwd()
    target_paths = [current_dir / t for t in targets]
    extra_stats = list(current_dir.glob("statistics_*.yaml"))
    target_paths.extend(extra_stats)


    if all_under_limit(target_paths):
        for path in target_paths:
            if path.exists():
                delete_path(path)
    else:
        print("❌ Deletion aborted: At least one target exceeds size limit.")


if __name__ == "__main__":
    main()



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################


import logging
from typing import Dict, List

import numpy as np
import torch
from ase import Atoms
from torch_geometric.loader import DataLoader

from .element import Element
from .quantity import KeySpecification
from ..utils.utils import log_statistics_to_yaml
from ..utils.torch_scatter import scatter, scatter_add


class OneHotToAtomicEnergy(torch.nn.Module):
    def __init__(self, atomic_energies: List[Dict[int, float]]) -> None:
        super().__init__()
        assert atomic_energies is not None
        atomic_energy_list = []
        for atomic_energy in atomic_energies:
            atomic_energy_list.append(
                [float(v) for _, v in atomic_energy.items()]
            )
        self.register_buffer(
            "atomic_energy",
            torch.tensor(
                atomic_energy_list,
                dtype=torch.get_default_dtype(),
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor: 
        return torch.matmul(x, self.atomic_energy.T)

    def __repr__(self):
        return f"{self.__class__.__name__}(atomic_eneries={[f'{x:.4f}' for x in self.atomic_energy.reshape(-1).tolist()]})"
    
class Statistics:
    """
    A container class for storing statistical information of physical quantities.

    This class dynamically accepts arbitrary keyword arguments and stores them as attributes.
    It is typically used to store mean, standard deviation, rms values, or other
    statistical metadata for normalization or scale-shift operations in the TACE framework.

    Attributes can be accessed directly using dot notation:

    Parameters
    ----------
    **kwargs : dict
        Arbitrary keyword arguments that will be stored as attributes of the instance.

    Examples
    --------
    >>> stats = Statistics(mean_energy=-5.2, std_force=0.03)
    >>> stats.mean_energy
    -5.2
    >>> 'std_force' in stats
    True
    """

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __contains__(self, key):
        return hasattr(self, key)

    def __repr__(self):
        attrs = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
        return f"Statistics({attrs})"

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __delitem__(self, key):
        delattr(self, key)

    def __iter__(self):
        return iter(self.__dict__)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def __repr__(self):
        attrs = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
        return f"Statistics({attrs})"


def _compute_statistics(
    dataloader_train: DataLoader,
    # dataloader_valid: DataLoader,
    atomic_numbers: List[str],
    atomic_energies: List[Dict[int, float]],
    target_property: List[str],
    device: str = "cpu",
    num_levels: int = 1,
) -> List[Statistics]:
    return [
            Statistics(**stat)
            for stat in analyze_dataset_statistics(
                dataloader_train,
                # dataloader_valid,
                atomic_numbers,
                atomic_energies,
                target_property,
                device,
                num_levels,
            )
        ]

def compute_atomic_energy_per_level(
    points: Atoms, element: Element, keyspec: KeySpecification,
) -> Dict[int, float]:
    len_train = len(points)
    A = np.zeros((len_train, len(element)))
    B = np.zeros(len_train)
    energyList = [
        points[i].info.get(keyspec.info_keys["energy"], None) for i in range(len_train)
    ]
    atomicNumbersList = [points[i].get_atomic_numbers() for i in range(len_train)]

    for i in range(len_train):
        B[i] = energyList[i]
        for j, z in enumerate(element.atomic_numbers):
            A[i, j] = np.count_nonzero(atomicNumbersList[i] == z)
    try:
        x = np.linalg.lstsq(A, B, rcond=None)[0]
        atomic_energy = {}
        for i, z in enumerate(element.atomic_numbers):
            atomic_energy[z] = float(x[i])
    except np.linalg.LinAlgError:
        logging.info(
            "Failed to compute Isolated Atomic Energies automatically , using Isolated Atomic Energies = 0.0 for all atoms"
        )
        atomic_energy = {}
        for i, z in enumerate(element.atomic_numbers):
            atomic_energy[z] = 0.0
    return atomic_energy

def compute_atomic_energy(
    points: Atoms, element: Element, keyspec: KeySpecification, num_levels: int,
) -> Dict[int, float]:
    atomic_energies = []
    len_train = len(points)
    for level_idx in range(num_levels):
        points_per_level = []
        for i in range(len_train):
            level = points[i].info.get(keyspec.info_keys['level'], 0)

            if level == level_idx:
                points_per_level.append(points[i])
        assert len(points_per_level) > 0
        atomic_energies.append(
            compute_atomic_energy_per_level(
                points_per_level, element, keyspec,
            )
        )
    return atomic_energies

def analyze_dataset_statistics(
    dataloader_train,
    # dataloader_valid,
    atomic_numbers: List[int],
    atomic_energies: List[Dict[int, float]],
    targe_property: List[str],
    device: str = "cpu",
    num_levels: int = 1,
) -> List[Dict]:

    energyList_per_level = [[] for _ in range(num_levels)]
    energyPerAtomList_per_level = [[] for _ in range(num_levels)]
    deltaEnergyPerAtomList_per_level = [[] for _ in range(num_levels)]
    forcesList_per_level = [[] for _ in range(num_levels)]
    elementIdxList_per_level = [[] for _ in range(num_levels)]

    neighborCountsList = []
    num_elements = len(atomic_numbers)
    neighbor_sum_per_element = torch.zeros(num_elements, device=device)
    count_per_element = torch.zeros(num_elements, dtype=torch.int64, device=device)

    compute_atomic_energy_fn = None
    if "energy" in targe_property:
        compute_atomic_energy_fn = OneHotToAtomicEnergy(atomic_energies).to(device)

    with torch.no_grad():
        for data in dataloader_train:
            data = data.to(device)
            num_graphs = len(data.ptr) - 1
            num_nodes = data.ptr[1:] - data.ptr[:-1]
            element_idx = data["node_attrs"].argmax(dim=-1)
            source = data.edge_index[0]

            # === average number of neighbors ===
            neighbor_counts = torch.bincount(source, minlength=data.batch.size(0)).float()
            neighborCountsList.append(neighbor_counts)
            neighbor_sum = scatter_add(
                neighbor_counts, element_idx, dim=0, dim_size=num_elements
            )
            element_count = scatter_add(
                torch.ones_like(neighbor_counts),
                element_idx,
                dim=0,
                dim_size=num_elements,
            )
            neighbor_sum_per_element += neighbor_sum
            count_per_element += element_count.long()

            if "level" in data:
                level = data["level"]
            else:
                level = torch.zeros(num_graphs, dtype=torch.int64, device=device)

            node_level = level[data['batch']]
            num_atoms_arange = torch.arange(
                data["node_attrs"].size(0), 
                device=data["node_attrs"].device,
                dtype=torch.int64,
            )
            # === energy ===
            if "energy" in targe_property:
                e0_node_energy = compute_atomic_energy_fn(
                    data["node_attrs"]
                )[num_atoms_arange, node_level]

                E0 = scatter(
                    e0_node_energy,
                    data['batch'],
                    dim=0,
                    dim_size=num_graphs,
                    reduce="sum",
                )
                energy = data['energy']
                energy_per_atom = energy / num_nodes
                delta_per_atom = (energy - E0) / num_nodes

                for lvl in range(num_levels):
                    mask = (level == lvl)
                    if mask.any():
                        energyList_per_level[lvl].append(energy[mask])
                        energyPerAtomList_per_level[lvl].append(energy_per_atom[mask])
                        deltaEnergyPerAtomList_per_level[lvl].append(delta_per_atom[mask])

            # === forces ===
            if "forces" in targe_property or "direct_forces" in targe_property:
                forces = data.get('forces', None)
                if forces is None and "direct_forces" in targe_property:
                    forces = data.get('direct_forces', None)
                for lvl in range(num_levels):
                    mask = (node_level == lvl)
                    if mask.any():
                        forcesList_per_level[lvl].append(forces[mask])
                        elementIdxList_per_level[lvl].append(element_idx[mask])

    # === Neighbor statistics ===
    neighborCounts = torch.cat(neighborCountsList)
    avg_num_neighbors = neighborCounts.mean().item()
    avg_neighbors_by_element = {
        atomic_numbers[i]: (
            (neighbor_sum_per_element[i] / count_per_element[i]).item()
            if count_per_element[i] > 0
            else 0.0
        )
        for i in range(num_elements)
    }

    def build_stats_from_lists(
        idx,
        energy_lists,
        energyPerAtom_lists,
        delta_lists,
        forces_lists,
        elemIdx_lists,
    ):
        stats = {
            "level": int(idx),
            "atomic_numbers": atomic_numbers,
            "avg_num_neighbors": avg_num_neighbors,
            "avg_neighbors_by_element": avg_neighbors_by_element,
        }

        # === Energy statistics === #
        if "energy" in targe_property:
            energy = torch.cat(energy_lists)
            energyPerAtom = torch.cat(energyPerAtom_lists)
            deltaEnergyPerAtom = torch.cat(delta_lists)
            mean_energy = energy.mean().item()
            std_energy = energy.std().item()
            mean_energy_per_atom = energyPerAtom.mean().item()
            mean_delta_energy_per_atom = deltaEnergyPerAtom.mean().item()
            stats.update(
                {
                    "__mean_energy": mean_energy,
                    "__std_energy": std_energy,
                    "__mean_energy_per_atom": mean_energy_per_atom,
                    "__mean_delta_energy_per_atom": mean_delta_energy_per_atom,
                    "atomic_energy": {k: float(v) for k, v in atomic_energies[idx].items()},
                    "scalar_mean_energy_per_atom": mean_energy_per_atom,
                    "mean_energy": {z: mean_energy for z in atomic_numbers},
                    "mean_energy_by_element": {z: mean_energy for z in atomic_numbers},
                    "std_energy": {z: std_energy for z in atomic_numbers},
                    "std_energy_by_element": {z: std_energy for z in atomic_numbers},
                    "mean_energy_per_atom": {
                        z: mean_energy_per_atom for z in atomic_numbers
                    },
                    "mean_energy_per_atom_by_element": {
                        z: mean_energy_per_atom for z in atomic_numbers
                    },
                    "mean_delta_energy_per_atom": {
                        z: mean_delta_energy_per_atom for z in atomic_numbers
                    },
                    "mean_delta_energy_per_atom_by_element": {
                        z: mean_delta_energy_per_atom for z in atomic_numbers
                    },
                }
            )

        # === Force statistics === #
        if "forces" in targe_property or 'direct_forces' in targe_property:
            forces = torch.cat(forces_lists, dim=0)  # [N, 3]
            elementIdx = torch.cat(elemIdx_lists, dim=0)  # [N]
            N = forces.size(0)

            # Global force statistics (3D)
            mean_forces_3d = forces.mean(dim=0).tolist()
            std_forces_3d = forces.std(dim=0).tolist()
            rms_forces_3d = torch.sqrt(torch.mean(forces**2, dim=0)).tolist()

            # Global force norm statistics (1D)
            forces_norm = torch.norm(forces, dim=1)  # [N]
            mean_forces_1d = forces_norm.mean().item()
            std_forces_1d = forces_norm.std().item()
            rms_forces_1d = torch.sqrt(torch.mean(forces_norm**2)).item()

            # Per-element 3D force statistics
            count = scatter_add(
                torch.ones((N, 1), device=device),
                elementIdx.unsqueeze(1),
                dim=0,
                dim_size=num_elements,
            ).clamp(min=1)
            mean_force = (
                scatter_add(forces, elementIdx, dim=0, dim_size=num_elements) / count
            )
            diff = forces - mean_force[elementIdx]
            std_force = torch.sqrt(
                scatter_add(diff**2, elementIdx, dim=0, dim_size=num_elements) / count
            )
            rms_force = torch.sqrt(
                scatter_add(forces**2, elementIdx, dim=0, dim_size=num_elements) / count
            )

            # Per-element 1D force norm statistics
            mean_force_1d_by_elem = (
                scatter_add(forces_norm, elementIdx, dim=0, dim_size=num_elements)
                / count.squeeze()
            )
            std_force_1d_by_elem = torch.sqrt(
                scatter_add(
                    (forces_norm - mean_force_1d_by_elem[elementIdx]) ** 2,
                    elementIdx,
                    dim=0,
                    dim_size=num_elements,
                )
                / count.squeeze()
            )
            rms_force_1d_by_elem = torch.sqrt(
                scatter_add(forces_norm**2, elementIdx, dim=0, dim_size=num_elements)
                / count.squeeze()
            )

            # Assemble results
            mean_forces_3d_by_element = {
                atomic_numbers[i]: mean_force[i].tolist() for i in range(num_elements)
            }
            std_forces_3d_by_element = {
                atomic_numbers[i]: std_force[i].tolist() for i in range(num_elements)
            }
            rms_forces_3d_by_element = {
                atomic_numbers[i]: rms_force[i].tolist() for i in range(num_elements)
            }
            mean_forces_1d_by_element = {
                atomic_numbers[i]: mean_force_1d_by_elem[i].item()
                for i in range(num_elements)
            }
            std_forces_1d_by_element = {
                atomic_numbers[i]: std_force_1d_by_elem[i].item()
                for i in range(num_elements)
            }
            rms_forces_1d_by_element = {
                atomic_numbers[i]: rms_force_1d_by_elem[i].item()
                for i in range(num_elements)
            }

            stats.update(
                {
                    # Global
                    "__mean_forces_3d": mean_forces_3d,
                    "__std_forces_3d": std_forces_3d,
                    "__rms_forces_3d": rms_forces_3d,
                    "__mean_forces_1d": mean_forces_1d,
                    "__std_forces_1d": std_forces_1d,
                    "__rms_forces_1d": rms_forces_1d,
                    # Global Per element
                    "__mean_forces_3d_by_element": mean_forces_3d_by_element,
                    "__std_forces_3d_by_element": std_forces_3d_by_element,
                    "__rms_forces_3d_by_element": rms_forces_3d_by_element,
                    "__mean_forces_1d_by_element": mean_forces_1d_by_element,
                    "__std_forces_1d_by_element": std_forces_1d_by_element,
                    "__rms_forces_1d_by_element": rms_forces_1d_by_element,
                    # for normalize:
                    "mean_forces_for_normalize": mean_forces_1d,
                    "std_forces_for_normalize": std_forces_1d,
                    # for scale
                    "rms_forces": {z: rms_forces_1d for z in atomic_numbers},
                    "std_forces": {z: std_forces_1d for z in atomic_numbers},
                    "rms_forces_by_element": rms_forces_1d_by_element,
                    "std_forces_by_element": std_forces_1d_by_element,
                }
            )

        return stats

    per_level_stats = []
    for lvl in range(num_levels):
        stats = build_stats_from_lists(
            lvl,
            energyList_per_level[lvl],
            energyPerAtomList_per_level[lvl],
            deltaEnergyPerAtomList_per_level[lvl],
            forcesList_per_level[lvl],
            elementIdxList_per_level[lvl],
        )
        per_level_stats.append(stats)

    log_statistics_to_yaml(per_level_stats)

    return per_level_stats




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import List, Dict


import torch
from torch import nn, Tensor




from .act import ACT
from .utils import expand_dims_to
from ...dataset.quantity import PROPERTY, UNIVERSAL_EMBEDDING_ALLOWED_PROPERTY


def add_rank1_to_left(T: Dict[int, torch.Tensor], rank1: torch.Tensor) -> Dict[int, torch.Tensor]:
    if 1 in T:
        T[1] = T[1] + rank1
    else:
        T[1] = rank1
    return T


def add_rank2_to_left(T: Dict[int, torch.Tensor], rank2: torch.Tensor) -> Dict[int, torch.Tensor]:
    if 2 in T:
        T[2] = T[2] + rank2
    else:
        T[2] = rank2
    return T


def add_rank3_to_left(T: Dict[int, torch.Tensor], rank3: torch.Tensor) -> Dict[int, torch.Tensor]:
    if 3 in T:
        T[3] = T[3] + rank3
    else:
        T[3] = rank3
    return T


ADD_FN = {
    1: add_rank1_to_left,
    2: add_rank2_to_left,
    3: add_rank3_to_left,
}


class UniversalInvariantEmbedding(torch.nn.Module):
    def __init__(
        self,
        out_dim: int,
        invariant_embedding: Dict[str, bool | str | int],
    ):
        super().__init__()

        self.embeddings = nn.ModuleDict()

        total_dim = 0
        for k, v in invariant_embedding.items():
            if v.get('enable', False) and PROPERTY[k]['rank'] == 0:
                p_type = PROPERTY[k]["type"]
                if p_type == "int":
                    self.embeddings[k] = nn.Embedding(v["num_embeddings"], out_dim)
                elif p_type == "float":
                    act = v.get("act", "silu")
                    self.embeddings[k] = nn.Sequential(
                        nn.Linear(1, out_dim, bias=False),
                        ACT[act](),
                        nn.Linear(out_dim, out_dim, bias=False),
                    )
                total_dim += out_dim

        self.project = nn.Sequential(
            nn.Linear(total_dim, out_dim, bias=False),
            nn.SiLU(),
        )

    def forward(
        self,
        batch: Tensor,
        attrs: Dict[str, Tensor],
    ) -> Tensor:
        embeddings = []

        for p, module in self.embeddings.items():
            attr = attrs[p]

            type_ = PROPERTY[p]['type']
            scope = PROPERTY[p]['scope']

            if scope == "per-system":
                attr = attr[batch]

            if type_ == 'float':
                attr = attr.unsqueeze(-1)
    
            embeddings.append(module(attr))

        return self.project(torch.cat(embeddings, dim=-1))


class EquivariantEmbedding(torch.nn.Module):
    def __init__(
        self,
        p: str,
        rank: int,
        scope: str,
        atomic_numbers: List,
        num_channel: int,
        element_trainable: bool = True,
        channel_trainable: bool = True,
        normalizer: float = 1.0
    ):
        super().__init__()
        num_elements = len(atomic_numbers)
        if element_trainable:
            self.element_weights = nn.Parameter(
                torch.ones(num_elements, dtype=torch.get_default_dtype())
            )
        else:
            self.register_buffer(
                "element_weights",
                torch.ones(num_elements, dtype=torch.get_default_dtype()),
            )
        if channel_trainable:
            self.channel_weights = nn.Parameter(
                torch.ones(num_channel, dtype=torch.get_default_dtype())
            )
        else:
            self.register_buffer(
                "channel_weights",
                torch.ones(num_channel, dtype=torch.get_default_dtype()),
            )
        self.p = p
        self.add_fn = ADD_FN[rank]
        self.scope = scope
        self.rank = rank
        self.normalizer = normalizer

    def forward(
        self,
        node_feats: Dict[int, Tensor],
        node_attrs: Tensor,
        batch: Tensor,
        attr: Dict[str, Tensor],
    ):
        element_idx = torch.argmax(node_attrs, dim=-1)
        label = attr * self.normalizer
        if self.scope == "per-system":
            label = label[batch].unsqueeze(1)
        else:
            label = label.unsqueeze(1)

        element_weights = expand_dims_to(
            self.element_weights[element_idx], n_dim=self.rank + 2
        )
        shape = (1, -1) + (1,) * self.rank
        channel_weights = self.channel_weights.view(*shape)
        
        embedding = label * element_weights * channel_weights
        return self.add_fn(node_feats, embedding)
  

class UniversalEquivariantEmbedding(torch.nn.Module):
    def __init__(
        self,
        equivariant_embedding: Dict[str, bool | str | int],
        atomic_numbers: List,
        num_channel: int,
    ):
        super().__init__()

        self.equivariant_embedding = equivariant_embedding
        self.embeddings = nn.ModuleDict()
        for k, v in equivariant_embedding.items():
            if v.get('enable', False) and PROPERTY[k]['rank'] > 0:
                self.embeddings[k] = EquivariantEmbedding(
                    k,
                    PROPERTY[k]["rank"],
                    PROPERTY[k]["scope"],
                    atomic_numbers,
                    num_channel,
                    element_trainable=True,
                    channel_trainable=True,
                    normalizer=float(v.get('normalizer', 1.0))
                )

    def forward(
            self, 
            node_feats: Dict[int, torch.Tensor], 
            node_attrs: torch.Tensor,
            batch: torch.Tensor,
            attrs: Dict[str, torch.Tensor]
        ) -> torch.Tensor:
        for p, module in self.embeddings.items():
            node_feats = module(node_feats, node_attrs, batch, attrs[p])
        return node_feats



from cartnn.o3._ictd import general_equivariant_basis_generator



space_1 = {"space":[[1,2],[1,3]],"parity":-1}
space_2 = {"space":[[1,3]],"parity":-1}

AS, BS = general_equivariant_basis_generator(space_1, space_2)


for A, B in zip(AS, BS):
    print(A)
    print(B.shape)


################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import numpy as np
from scipy.fftpack import dct


from .units import c, hbar, Boltzmann


def ACF(x):
    n = len(x)
    return np.correlate(x, x, "full")[-n:] / np.arange(n, 0, -1)


def d_ACF(dipole, N: int):
    dacf = ACF(dipole[:, 0]) + ACF(dipole[:, 1]) + ACF(dipole[:, 2])
    dacf = dacf[:N]
    return dacf


def pGamma_ACF(polar, N: int):
    gamma = np.zeros(len(polar))
    for i in range(len(polar)):
        gamma[i] = np.trace(polar[i]) / 3
    pgammaacf = ACF(gamma)
    pgammaacf = pgammaacf[:N]
    return pgammaacf


def pBeta_ACF(polar, N: int):
    beta = np.zeros_like(polar)
    for i in range(len(polar)):
        beta[i] = polar[i] - np.eye(3) * np.trace(polar[i]) / 3
    pbetaacf = np.zeros(len(beta))
    for i in range(3):
        for j in range(3):
            pbetaacf += ACF(beta[:, i, j])
    pbetaacf = pbetaacf[:N]
    return pbetaacf


def IR(dacf, N: int, T: float, dt: float, w_max: float):
    ir = dct(dacf, type=1)
    freq = np.linspace(0, 0.5 / dt, N) / (100.0 * c)
    ir = freq**2 * ir
    n = int(w_max / freq.max() * N)
    return freq[:n], ir[:n]


def Raman(acf_beta, N, T, dt, w_max):
    raman = dct(acf_beta, type=1)
    freq = np.linspace(0, 0.5 / dt, N) / (100.0 * c)
    raman = freq**2 * raman
    n = int(w_max / freq.max() * N)
    return freq[:n], raman[:n]
